#
#  Copyright (C) 2022 Novartis Institute of BioMedical Research
#  and other RDKit contributors
#
#   @@ All Rights Reserved @@
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.
#
""" Interactive molecule rendering through rdkit-structure-renderer.js """
import base64
import json
import logging
import re
import uuid
from xml.dom import minidom

from IPython.display import HTML, display

from rdkit import Chem
from rdkit.Chem import Draw

from . import rdMolDraw2D

log = logging.getLogger(__name__)

rdkitStructureRendererJsUrl = "https://unpkg.com/rdkit-structure-renderer/dist/rdkit-structure-renderer-module.js"
minimalLibJsUrl = "https://unpkg.com/rdkit-structure-renderer/public/RDKit_minimal.js"
parentNodeQuery = "div[class*=jp-NotebookPanel-notebook]"
_enabled_div_uuid = None
_camelCaseOptToTagRe = re.compile("[A-Z]")
_opts = "__rnrOpts"
_rdk_str_rnr_prefix = "rdk-str-rnr-"
_disabled = "_disabled"
_defaultDrawOptions = None
_defaultDrawOptionsDict = None


def _isAcceptedKeyValue(key, value):
  return not key.startswith("__") and (type(value) in (bool, int, float, str, tuple, list) or
                                       (callable(value) and key.startswith("get")))


def _getValueFromKey(d, key):
  value = getattr(d, key)
  return value() if callable(value) else value


def _molDrawOptionsToDict(molDrawOptions=rdMolDraw2D.MolDrawOptions()):
  return {
    key: _getValueFromKey(molDrawOptions, key)
    for key in dir(molDrawOptions) if _isAcceptedKeyValue(key, getattr(molDrawOptions, key))
  }


def _isDrawOptionEqual(v1, v2):
  if type(v1) != type(v2):
    return False
  if type(v1) in (tuple, list):
    if len(v1) != len(v2):
      return False
    return all(_isDrawOptionEqual(item1, v2[i]) for i, item1 in enumerate(v1))
  if type(v1) == float:
    return abs(v1 - v2) < 1.e-5
  return v1 == v2


def filterDefaultDrawOpts(molDrawOptions):
  global _defaultDrawOptionsDict
  if not isinstance(molDrawOptions, rdMolDraw2D.MolDrawOptions):
    raise ValueError(f"Bad args ({str(type(molDrawOptions))}) for {__name__}.filterDefaultDrawOpts("
                     "molDrawOptions: Chem.Draw.rdMolDraw2D.MolDrawOptions)")
  molDrawOptionsAsDict = _molDrawOptionsToDict(molDrawOptions)
  if (_defaultDrawOptionsDict is None):
    _defaultDrawOptionsDict = _molDrawOptionsToDict()
  return {
    key: value
    for key, value in molDrawOptionsAsDict.items()
    if not _isDrawOptionEqual(value, _defaultDrawOptionsDict[key])
  }


def setEnabled(shouldEnable=True, quiet=False):
  """ Enable interactive molecule rendering """

  def _wrapMsgIntoDiv(uuid, msg, quiet, script=None):
    return ((script or "") + '<div '
            'class="lm-Widget p-Widget jp-RenderedText jp-mod-trusted jp-OutputArea-output"'
            f'id="{uuid}">{"" if quiet else msg}</div>')

  global _enabled_div_uuid
  loadingMsg = "Loading rdkit-structure-renderer.js..."
  failedToLoadMsg = "Failed to load rdkit-structure-renderer.js:"
  renderingMsg = "Interactive molecule rendering is {0:s} in this Jupyter Notebook."
  renderingUnavailableMsg = renderingMsg.format("not available")
  renderingEnabledMsg = renderingMsg.format("enabled")
  renderingDisabledMsg = renderingMsg.format("disabled")
  checkRenderingAvailable = f"""<script>\
(() => {{
  const rdkStrRnr = window.rdkStrRnr || Promise.resolve(null);
  if (rdkStrRnr) {{
    rdkStrRnr.then(
      (Renderer) => {{
        if (Renderer && Renderer.updateMolDrawDivs) {{
          setTimeout(() => {{
            const jsLoader = document.getElementById('{_enabled_div_uuid}') || {{}};
            jsLoader.innerHTML = '{renderingEnabledMsg}';
          }}, 10);
        }} else {{
          window.rdkStrRnr = null;
        }}
      }}
    ).catch((e) => {{
      console.error(e.toString());
    }});
}})();
}}</script>"""
  if not shouldEnable:
    _enabled_div_uuid = None
    return display(HTML(_wrapMsgIntoDiv("", renderingDisabledMsg, quiet)))
  if _enabled_div_uuid:
    return display(HTML(_wrapMsgIntoDiv(_enabled_div_uuid, renderingUnavailableMsg, quiet, script=checkRenderingAvailable)))
  _enabled_div_uuid = str(uuid.uuid1())
  return display(
    HTML(
      _wrapMsgIntoDiv(_enabled_div_uuid, loadingMsg, quiet) + f"""<script type="module">
const jsLoader = document.getElementById('{_enabled_div_uuid}') || {{}};
const setError = (e) => {{
  jsLoader.innerHTML = (
    '{failedToLoadMsg}<br>' +
    e.toString() + '<br>' +
    '{renderingUnavailableMsg}'
  );
  window.rdkStrRnr = null;
}};
if (window.rdkStrRnr) {{
  jsLoader.innerHTML = '{renderingEnabledMsg}';
}} else {{
  window.rdkStrRnr = new Promise((resolve) => {{
    try {{
      fetch('{rdkitStructureRendererJsUrl}').then(
        (r) => r.text().then(
          (t) => import(URL.createObjectURL(new Blob([t], {{ type: 'application/javascript' }}))).then(
            ({{ default: Renderer }}) => {{
              const res = Renderer.init('{minimalLibJsUrl}');
              return res.then(
                (Renderer) => {{
                  jsLoader.innerHTML = '{renderingEnabledMsg}';
                  resolve(Renderer);
                }}
              ).catch(setError);
            }}
          ).catch(setError)
        ).catch(setError)
      ).catch(setError);
    }} catch(e) {{
      setError(e);
    }}
  }});
}}
</script>"""))


def isEnabled(mol=None):
  return _enabled_div_uuid and (mol is None or not isNoInteractive(mol))


def getOpts(mol):
  if not isinstance(mol, Chem.Mol):
    raise ValueError(f"Bad args ({str(type(mol))}) for {__name__}.getOpts(mol: Chem.Mol)")
  opts = {}
  if hasattr(mol, _opts):
    opts = getattr(mol, _opts)
    if not isinstance(opts, dict):
      opts = {}
  return opts


def setOpts(mol, opts):
  if not isinstance(mol, Chem.Mol) or not isinstance(opts, dict):
    raise ValueError(
      f"Bad args ({str(type(mol)), str(type(opts))}) for {__name__}.setOpts(mol: Chem.Mol, opts: dict)"
    )
  if not all(opts.keys()):
    raise ValueError(
      f"{__name__}.setOpts(mol: Chem.Mol, opts: dict): no key in opts should be null")
  if opts:
    setattr(mol, _opts, opts)
  elif hasattr(mol, _opts):
    delattr(mol, _opts)


def setOpt(mol, key, value):
  if not isinstance(mol, Chem.Mol) or not isinstance(key, str) or not key:
    raise ValueError(
      f"Bad args ({str(type(mol))}, {str(type(key))}) for {__name__}.setOpt(mol: Chem.Mol, key: str, value: Any)"
    )
  opts = getOpts(mol)
  opts[key] = value
  setOpts(mol, opts)


def clearOpts(mol):
  if not isinstance(mol, Chem.Mol):
    raise ValueError(f"Bad args ({str(type(mol))}) for {__name__}.clearOpts(mol: Chem.Mol)")
  setOpts(mol, {})


def clearOpt(mol, key):
  if not isinstance(mol, Chem.Mol) or not isinstance(key, str):
    raise ValueError(
      f"Bad args ({str(type(mol))}, {str(type(key))}) for {__name__}.clearOpt(mol: Chem.Mol, key: str)"
    )
  opts = getOpts(mol)
  if key in opts:
    opts.pop(key)
  setOpts(mol, opts)


def setNoInteractive(mol, shouldSet=True):
  opts = getOpts(mol)
  if shouldSet:
    opts[_disabled] = True
  elif _disabled in opts:
    opts.pop(_disabled)
  setOpts(mol, opts)


def isNoInteractive(mol):
  opts = getOpts(mol)
  return _disabled in opts


def injectHTMLFooterAfterTable(html):
  doc = minidom.parseString(html.replace(" scoped", ""))
  tables = doc.getElementsByTagName("table")
  for table in tables:
    tbody = table.getElementsByTagName("tbody")
    if tbody:
      if len(tbody) != 1:
        return html
      tbody = tbody.pop(0)
    else:
      tbody = table
    div_list = tbody.getElementsByTagName("div")
    if any(div.getAttribute("class") == f"{_rdk_str_rnr_prefix}mol-container" for div in div_list):
      return generateHTMLFooter(doc, table)
  return html


def generateHTMLFooter(doc, element):
  element_parent = element.parentNode
  if element_parent.nodeName.lower() != "div":
    element_parent = doc.createElement("div")
    element_parent.appendChild(element)
  script = doc.createElement("script")
  script.setAttribute("type", "module")
  # avoid arrow function as minidom encodes => into HTML (grrr)
  # Also use single quotes to avoid the &quot; encoding
  cmd = doc.createTextNode(f"""\
if (window.rdkStrRnr) {{
  window.rdkStrRnr.then(
    function(Renderer) {{
      if (Renderer) {{
        Renderer.updateMolDrawDivs();
      }}
    }}
  ).catch();
}}""")
  script.appendChild(cmd)
  element_parent.appendChild(script)
  html = element_parent.toxml()
  return html


def newlineToXml(molblock):
  return molblock.replace("\n", "<\\n>")


def xmlToNewline(xmlblock):
  return xmlblock.replace("&lt;\\n&gt;", "&#10;")


def toDataMol(mol):
  return "pkl_" + base64.b64encode(
    mol.ToBinary(Chem.PropertyPickleOptions.AllProps
                 ^ Chem.PropertyPickleOptions.ComputedProps)).decode("utf-8")


def _dashLower(m):
  return "-" + m.group(0).lower()


def camelCaseOptToDataTag(opt):
  tag = _camelCaseOptToTagRe.sub(_dashLower, opt)
  if not tag.startswith("data-"):
    tag = "data-" + tag
  return tag


def generateHTMLBody(mol, size, **kwargs):

  def toJson(x):
    return json.dumps(x, separators=(",", ":"))

  drawOptions = kwargs.get("drawOptions", _defaultDrawOptions)
  legend = kwargs.get("legend", None)
  useSVG = kwargs.get("useSVG", False)
  kekulize = Draw.shouldKekulize(mol, kwargs.get("kekulize", True))
  highlightAtoms = kwargs.get("highlightAtoms", []) or []
  highlightBonds = kwargs.get("highlightBonds", []) or []
  if not highlightAtoms and hasattr(mol, '__sssAtoms'):
    highlightAtoms = mol.__sssAtoms
    highlightBonds = [
      b.GetIdx() for b in mol.GetBonds()
      if b.GetBeginAtomIdx() in highlightAtoms and b.GetEndAtomIdx() in highlightAtoms
    ]
  userDrawOpts = filterDefaultDrawOpts(drawOptions)
  molOpts = getOpts(mol)
  molOptsDashed = {}
  for key, value in molOpts.items():
    keyDashed = camelCaseOptToDataTag(key)
    if keyDashed == "data-draw-opts":
      if isinstance(value, str):
        value = json.loads(value)
      if not isinstance(value, dict):
        raise ValueError(f"data-draw-opts: expected dict, found {str(type(value))}")
      userDrawOpts.update(value)
    else:
      molOptsDashed[keyDashed] = value
  doc = minidom.Document()
  unique_id = str(uuid.uuid1())
  div = doc.createElement("div")
  for key, value in [
    ("style", f"margin: auto;"),
    ("class", f"{_rdk_str_rnr_prefix}mol-container"),
    ("id", f"{_rdk_str_rnr_prefix}mol-{unique_id}"),
    ("data-mol", toDataMol(mol)),
    ("data-content", "rdkit/molecule"),
    ("data-parent-node", parentNodeQuery),
    ("data-width", str(size[0])),
    ("data-height", str(size[1])),
  ]:
    div.setAttribute(key, value)
  if "addAtomIndices" in userDrawOpts:
    addAtomIndices = userDrawOpts["addAtomIndices"]
    if addAtomIndices:
      molOptsDashed["data-atom-idx"] = True
    userDrawOpts.pop("addAtomIndices")
  if highlightAtoms or highlightBonds:
    molOptsDashed["data-scaffold-highlight"] = True
    userDrawOpts["atoms"] = highlightAtoms
    userDrawOpts["bonds"] = highlightBonds
  if not kekulize:
    userDrawOpts["kekulize"] = False
  for key, value in molOptsDashed.items():
    if isinstance(value, Chem.Mol):
      value = toDataMol(value)
    elif not any(isinstance(value, t) for t in (str, int, float)):
      value = toJson(value)
    div.setAttribute(key, str(value))
  if userDrawOpts:
    div.setAttribute("data-draw-opts", toJson(userDrawOpts))
  if useSVG:
    div.setAttribute("data-use-svg", "true")
  if legend:
    outerTable = doc.createElement("table")
    outerTable.setAttribute("style", f"margin: auto;")
    molTr = doc.createElement("tr")
    molTd = doc.createElement("td")
    molTd.setAttribute("style", f"padding: 0;")
    molTd.appendChild(div)
    molTr.appendChild(molTd)
    nameTr = doc.createElement("tr")
    nameTh = doc.createElement("th")
    legendText = doc.createTextNode(legend)
    nameTh.appendChild(legendText)
    nameTh.setAttribute("style", f"text-align: center; background: white;")
    nameTr.appendChild(nameTh)
    outerTable.appendChild(molTr)
    outerTable.appendChild(nameTr)
    div = outerTable
  return xmlToNewline(div.toxml())


def MolsToHTMLTable(mols, molsPerRow=3, subImgSize=(200, 200), legends=None,
                    highlightAtomLists=None, highlightBondLists=None, useSVG=False,
                    drawOptions=None, **kwargs):
  if legends and len(legends) > len(mols):
    legends = legends[:len(mols)]
  if highlightAtomLists and len(highlightAtomLists) > len(mols):
    highlightAtomLists = highlightAtomLists[:len(mols)]
  if highlightBondLists and len(highlightBondLists) > len(mols):
    highlightBondLists = highlightBondLists[:len(mols)]
  nRows = (len(mols) - 1) // molsPerRow + 1 if mols else 0
  if nRows:
    doc = minidom.Document()
    table = doc.createElement("table")
  res = ""
  i = 0
  for _ in range(nRows):
    tr = doc.createElement("tr")
    for _ in range(molsPerRow):
      td = doc.createElement("td")
      td.setAttribute("style", f"padding: 0; background-color: white;")
      highlights = None
      mol = mols[i]
      legend = legends[i] if legends else None
      highlights = highlightAtomLists[i] if highlightAtomLists else None
      kwargs["highlightBonds"] = highlightBondLists[i] if highlightBondLists else None
      content = None
      if isinstance(mol, Chem.Mol):
        if isEnabled(mol):
          content = generateHTMLBody(mol, subImgSize, drawOptions=drawOptions, legend=legend,
                                     useSVG=useSVG, highlightAtoms=highlights, **kwargs)
        else:
          fn = Draw._moltoSVG if useSVG else Draw._moltoimg
          content = fn(mol, subImgSize, highlights, legend, **kwargs)
      try:
        content = minidom.parseString(content)
        td.appendChild(content.firstChild)
      except Exception as e:
        log.warning("Failed to parse HTML returned by generateHTMLBody()")
      tr.appendChild(td)
      i += 1
      if i == len(mols):
        break
    table.appendChild(tr)
  if nRows:
    doc.appendChild(table)
    res = doc.toxml()
  return injectHTMLFooterAfterTable(res)



#
#  Copyright (c) 2011, Novartis Institutes for BioMedical Research Inc.
#  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#     * Neither the name of Novartis Institutes for BioMedical Research Inc.
#       nor the names of its contributors may be used to endorse or promote
#       products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

INCHI_AVAILABLE = True

import logging
import re

from rdkit import Chem, Geometry
from rdkit import RDLogger
from rdkit.Chem import rdinchi

logger = RDLogger.logger()

logLevelToLogFunctionLookup = {
  logging.INFO: logger.info,
  logging.DEBUG: logger.debug,
  logging.WARNING: logger.warning,
  logging.CRITICAL: logger.critical,
  logging.ERROR: logger.error
}


class InchiReadWriteError(Exception):
  pass


def _parse_auxinfo_coordinates(auxinfo):
  """Parse the rC: (coordinate) layer from an InChI AuxInfo string.

  Returns (coords_list, is_3d) where coords_list is a list of (x, y, z) tuples
  in original input atom order, or (None, None) if parsing fails or coords are empty.
  """
  if not auxinfo:
    return None, None
  match = re.search(r'/rC:([^/]+)', auxinfo)
  if not match:
    return None, None
  entries = match.group(1).split(';')
  # Remove trailing empty entries from trailing semicolons
  while entries and not entries[-1].strip():
    entries.pop()
  coords = []
  for entry in entries:
    entry = entry.strip()
    if not entry:
      return None, None
    parts = entry.split(',')
    try:
      x = float(parts[0]) if parts[0] else 0.0
      y = float(parts[1]) if len(parts) > 1 and parts[1] else 0.0
      z = float(parts[2]) if len(parts) > 2 and parts[2] else 0.0
    except (ValueError, IndexError):
      return None, None
    coords.append((x, y, z))
  if not coords:
    return None, None
  # All-zero coordinates means no real coords were present
  if all(x == 0.0 and y == 0.0 and z == 0.0 for x, y, z in coords):
    return None, None
  is_3d = any(z != 0.0 for _, _, z in coords)
  return coords, is_3d


def _parse_auxinfo_atom_order(auxinfo):
  """Parse the N: (atom numbering) layer from an InChI AuxInfo string.

  Returns a list of 0-based original atom indices, or None if parsing fails.
  The returned list maps from InChI canonical order to original atom order:
  result[i] is the original atom index for InChI canonical atom i.
  """
  if not auxinfo:
    return None
  match = re.search(r'/N:([^/]+)', auxinfo)
  if not match:
    return None
  # The N: layer contains comma-separated 1-based atom indices
  # possibly with semicolons separating disconnected fragments
  tokens = match.group(1).replace(';', ',').split(',')
  try:
    return [int(t) - 1 for t in tokens]
  except (ValueError, IndexError):
    return None


def _attach_conformer(mol, coords, is_3d):
  """Attach parsed /rC: coordinates to a molecule as a conformer."""
  if coords is not None and len(coords) == mol.GetNumAtoms():
    conf = Chem.Conformer(mol.GetNumAtoms())
    conf.Set3D(is_3d)
    for i, (x, y, z) in enumerate(coords):
      conf.SetAtomPosition(i, Geometry.Point3D(x, y, z))
    mol.AddConformer(conf, assignId=True)
  return mol


def _build_inverse_permutation(atom_order, size):
  """Build the inverse permutation for RenumberAtoms.

  atom_order[inchi_idx] = original_idx. Returns new_order where
  new_order[original_idx] = inchi_idx, or None if any index is out of range.
  """
  new_order = [0] * size
  for inchi_idx, orig_idx in enumerate(atom_order):
    if orig_idx >= size:
      return None
    new_order[orig_idx] = inchi_idx
  return new_order


def MolFromInchiAndAuxInfo(inchi, auxinfo, sanitize=True, removeHs=True, logLevel=None,
                           treatWarningAsError=False):
  """Construct a molecule from an InChI string and its AuxInfo, restoring the
  original atom ordering.

    Keyword arguments:
    sanitize -- set to True to enable sanitization of the molecule. Default is
    True
    removeHs -- set to True to remove Hydrogens from a molecule. This only
    makes sense when sanitization is enabled
    logLevel -- the log level used for logging logs and messages from InChI
    API. set to None to diable the logging completely
    treatWarningAsError -- set to True to raise an exception in case of a
    molecule that generates warning in calling InChI API. The resultant
    molecule and error message are part of the excpetion

    Returns:
    a rdkit.Chem.rdchem.Mol instance with atoms reordered to match the
    original atom ordering encoded in the AuxInfo
    """
  mol = MolFromInchi(inchi, sanitize=sanitize, removeHs=removeHs, logLevel=logLevel,
                     treatWarningAsError=treatWarningAsError)
  if mol is None:
    return None

  # /rC: coordinates are in original input order; attach after reordering.
  coords, is_3d = _parse_auxinfo_coordinates(auxinfo)

  atom_order = _parse_auxinfo_atom_order(auxinfo)
  if atom_order is None:
    return _attach_conformer(mol, coords, is_3d)

  from rdkit.Chem import RenumberAtoms
  num_mol_atoms = mol.GetNumAtoms()

  if removeHs:
    # N: layer lists heavy atoms only; must match molecule size after H removal
    if len(atom_order) != num_mol_atoms:
      return mol
    new_order = _build_inverse_permutation(atom_order, num_mol_atoms)
    if new_order is None:
      return mol
    return _attach_conformer(RenumberAtoms(mol, new_order), coords, is_3d)

  if len(atom_order) > num_mol_atoms:
    return mol

  if len(atom_order) == num_mol_atoms:
    new_order = _build_inverse_permutation(atom_order, num_mol_atoms)
    if new_order is None:
      return mol
    return _attach_conformer(RenumberAtoms(mol, new_order), coords, is_3d)

  # More atoms in molecule than in atom_order (explicit Hs added by InChI).
  # Place heavy atoms in original order, then append Hs.
  num_heavy = len(atom_order)
  new_order = _build_inverse_permutation(atom_order, num_heavy)
  if new_order is None:
    return mol
  new_order.extend([0] * (num_mol_atoms - num_heavy))
  h_slot = num_heavy
  for i in range(num_mol_atoms):
    if i not in atom_order:
      if h_slot >= num_mol_atoms:
        return mol
      new_order[h_slot] = i
      h_slot += 1
  return _attach_conformer(RenumberAtoms(mol, new_order), coords, is_3d)


def MolFromInchi(inchi, sanitize=True, removeHs=True, logLevel=None, treatWarningAsError=False):
  """Construct a molecule from a InChI string

    Keyword arguments:
    sanitize -- set to True to enable sanitization of the molecule. Default is
    True
    removeHs -- set to True to remove Hydrogens from a molecule. This only
    makes sense when sanitization is enabled
    logLevel -- the log level used for logging logs and messages from InChI
    API. set to None to diable the logging completely
    treatWarningAsError -- set to True to raise an exception in case of a
    molecule that generates warning in calling InChI API. The resultant
    molecule  and error message are part of the excpetion

    Returns:
    a rdkit.Chem.rdchem.Mol instance
    """
  try:
    mol, retcode, message, log = rdinchi.InchiToMol(inchi, sanitize, removeHs)
  except ValueError as e:
    logger.error(str(e))
    return None

  if logLevel is not None:
    if logLevel not in logLevelToLogFunctionLookup:
      raise ValueError("Unsupported log level: %d" % logLevel)
    log = logLevelToLogFunctionLookup[logLevel]
    if retcode == 0:
      log(message)

  if retcode != 0:
    if retcode == 1:
      logger.warning(message)
    else:
      logger.error(message)
  if treatWarningAsError and retcode != 0:
    raise InchiReadWriteError(mol, message)
  return mol


def MolToInchiAndAuxInfo(mol, options="", logLevel=None, treatWarningAsError=False):
  """Returns the standard InChI string and InChI auxInfo for a molecule

    Keyword arguments:
    logLevel -- the log level used for logging logs and messages from InChI
    API. set to None to diable the logging completely
    treatWarningAsError -- set to True to raise an exception in case of a
    molecule that generates warning in calling InChI API. The resultant InChI
    string and AuxInfo string as well as the error message are encoded in the
    exception.

    Returns:
    a tuple of the standard InChI string and the auxInfo string returned by
    InChI API, in that order, for the input molecule
    """
  inchi, retcode, message, logs, aux = rdinchi.MolToInchi(mol, options)
  if logLevel is not None:
    if logLevel not in logLevelToLogFunctionLookup:
      raise ValueError("Unsupported log level: %d" % logLevel)
    log = logLevelToLogFunctionLookup[logLevel]
    if retcode == 0:
      log(message)
  if retcode != 0:
    if retcode == 1:
      logger.warning(message)
    else:
      logger.error(message)

  if treatWarningAsError and retcode != 0:
    raise InchiReadWriteError(inchi, aux, message)
  return inchi, aux


def MolBlockToInchiAndAuxInfo(molblock, options="", logLevel=None, treatWarningAsError=False):
  """Returns the standard InChI string and InChI auxInfo for a mol block

    Keyword arguments:
    logLevel -- the log level used for logging logs and messages from InChI
    API. set to None to diable the logging completely
    treatWarningAsError -- set to True to raise an exception in case of a
    molecule that generates warning in calling InChI API. The resultant InChI
    string and AuxInfo string as well as the error message are encoded in the
    exception.

    Returns:
    a tuple of the standard InChI string and the auxInfo string returned by
    InChI API, in that order, for the input molecule
    """
  inchi, retcode, message, logs, aux = rdinchi.MolBlockToInchi(molblock, options)
  if logLevel is not None:
    if logLevel not in logLevelToLogFunctionLookup:
      raise ValueError("Unsupported log level: %d" % logLevel)
    log = logLevelToLogFunctionLookup[logLevel]
    if retcode == 0:
      log(message)
  if retcode != 0:
    if retcode == 1:
      logger.warning(message)
    else:
      logger.error(message)

  if treatWarningAsError and retcode != 0:
    raise InchiReadWriteError(inchi, aux, message)
  return inchi, aux


def MolToInchi(mol, options="", logLevel=None, treatWarningAsError=False):
  """Returns the standard InChI string for a molecule

    Keyword arguments:
    logLevel -- the log level used for logging logs and messages from InChI
    API. set to None to diable the logging completely
    treatWarningAsError -- set to True to raise an exception in case of a
    molecule that generates warning in calling InChI API. The resultant InChI
    string and AuxInfo string as well as the error message are encoded in the
    exception.

    Returns:
    the standard InChI string returned by InChI API for the input molecule
    """
  if options.find('AuxNone') == -1:
    if options:
      options += " /AuxNone"
    else:
      options += "/AuxNone"

  try:
    inchi, aux = MolToInchiAndAuxInfo(mol, options, logLevel=logLevel,
                                      treatWarningAsError=treatWarningAsError)
  except InchiReadWriteError as inst:
    inchi, aux, message = inst.args
    raise InchiReadWriteError(inchi, message)
  return inchi


def MolBlockToInchi(molblock, options="", logLevel=None, treatWarningAsError=False):
  """Returns the standard InChI string for a mol block

    Keyword arguments:
    logLevel -- the log level used for logging logs and messages from InChI
    API. set to None to diable the logging completely
    treatWarningAsError -- set to True to raise an exception in case of a
    molecule that generates warning in calling InChI API. The resultant InChI
    string and AuxInfo string as well as the error message are encoded in the
    exception.

    Returns:
    the standard InChI string returned by InChI API for the input molecule
    """
  if options.find('AuxNone') == -1:
    if options:
      options += " /AuxNone"
    else:
      options += "/AuxNone"

  try:
    inchi, aux = MolBlockToInchiAndAuxInfo(molblock, options, logLevel=logLevel,
                                           treatWarningAsError=treatWarningAsError)
  except InchiReadWriteError as inst:
    inchi, aux, message = inst.args
    raise InchiReadWriteError(inchi, message)
  return inchi


def InchiToInchiKey(inchi):
  """Return the InChI key for the given InChI string. Return None on error"""
  ret = rdinchi.InchiToInchiKey(inchi)
  if ret:
    return ret
  else:
    return None


def MolToInchiKey(mol, options=""):
  """Returns the standard InChI key for a molecule

    Returns:
    the standard InChI key returned by InChI API for the input molecule
    """
  return rdinchi.MolToInchiKey(mol, options)


GetInchiVersion = rdinchi.GetInchiVersion

__all__ = [
  'MolToInchiAndAuxInfo', 'MolToInchi', 'MolBlockToInchiAndAuxInfo', 'MolBlockToInchi',
  'MolFromInchi', 'MolFromInchiAndAuxInfo', 'InchiReadWriteError', 'InchiToInchiKey',
  'MolToInchiKey', 'GetInchiVersion', 'INCHI_AVAILABLE'
]



#
#  Copyright (c) 2016, Novartis Institutes for BioMedical Research Inc.
#  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#     * Neither the name of Novartis Institutes for BioMedical Research Inc.
#       nor the names of its contributors may be used to endorse or promote
#       products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Created by Nadine Schneider, July 2016

import copy
from collections import defaultdict

from rdkit import Chem
from rdkit.Chem import AllChem


def transferAgentsToReactants(rxn):
  for a in range(rxn.GetNumAgentTemplates()):
    agent = rxn.GetAgentTemplate(a)
    rxn.AddReactantTemplate(agent)


def removeAgentsAndTransferToReactants(rxn):
  tmp = []
  rxn.RemoveAgentTemplates(tmp)
  for a in tmp:
    rxn.AddReactantTemplate(a)


def getNumPositiveCounts(fp):
  count = 0
  for k, v in fp.GetNonzeroElements().items():
    if v > 0:
      count += v
  return count


def getNumNegativeCounts(fp):
  count = 0
  for k, v in fp.GetNonzeroElements().items():
    if v < 0:
      count += abs(v)
  return count


def getNumPositiveBitCountsOfRadius0(fp, bitinfo):
  count = 0
  bitsUnmappedAtoms = []
  for k in bitinfo:
    if bitinfo[k][0][1] == 0:
      v = fp[k]
      if v > 0:
        count += 1
        bitsUnmappedAtoms.append((k, v))
  return count, bitsUnmappedAtoms


def getSumFps(fps):
  summedFP = copy.deepcopy(fps[0])
  for fp in fps[1:]:
    summedFP += fp
  return summedFP


def uniqueMolecules(mols):
  smiles = [Chem.MolToSmiles(mol) for mol in mols]
  uniqueMolecules = defaultdict(int)
  for n, smi in enumerate(smiles):
    uniqueMolecules[n] = smiles.index(smi)
  return uniqueMolecules, smiles



#  Copyright (C) 2005-2006 greg landrum and Rational Discovery LLC
#
#   @@ All Rights Reserved @@
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.
#

import copy
import struct

from rdkit import DataStructs


class VectCollection(object):
  """

    >>> vc = VectCollection()
    >>> bv1 = DataStructs.ExplicitBitVect(10)
    >>> bv1.SetBitsFromList((1,3,5))
    >>> vc.AddVect(1,bv1)
    >>> bv1 = DataStructs.ExplicitBitVect(10)
    >>> bv1.SetBitsFromList((6,8))
    >>> vc.AddVect(2,bv1)
    >>> len(vc)
    10
    >>> vc.GetNumBits()
    10
    >>> vc[0]
    0
    >>> vc[1]
    1
    >>> vc[9]
    0
    >>> vc[6]
    1
    >>> vc.GetBit(6)
    1
    >>> list(vc.GetOnBits())
    [1, 3, 5, 6, 8]

    keys must be unique, so adding a duplicate replaces the
    previous values:

    >>> bv1 = DataStructs.ExplicitBitVect(10)
    >>> bv1.SetBitsFromList((7,9))
    >>> vc.AddVect(1,bv1)
    >>> len(vc)
    10
    >>> vc[1]
    0
    >>> vc[9]
    1
    >>> vc[6]
    1

    we can also query the children:

    >>> vc.NumChildren()
    2
    >>> cs = vc.GetChildren()
    >>> id,fp = cs[0]
    >>> id
    1
    >>> list(fp.GetOnBits())
    [7, 9]
    >>> id,fp = cs[1]
    >>> id
    2
    >>> list(fp.GetOnBits())
    [6, 8]

    attach/detach operations:

    >>> bv1 = DataStructs.ExplicitBitVect(10)
    >>> bv1.SetBitsFromList((5,6))
    >>> vc.AddVect(3,bv1)
    >>> vc.NumChildren()
    3
    >>> list(vc.GetOnBits())
    [5, 6, 7, 8, 9]
    >>> vc.DetachVectsNotMatchingBit(6)
    >>> vc.NumChildren()
    2
    >>> list(vc.GetOnBits())
    [5, 6, 8]


    >>> bv1 = DataStructs.ExplicitBitVect(10)
    >>> bv1.SetBitsFromList((7,9))
    >>> vc.AddVect(1,bv1)
    >>> vc.NumChildren()
    3
    >>> list(vc.GetOnBits())
    [5, 6, 7, 8, 9]
    >>> vc.DetachVectsMatchingBit(6)
    >>> vc.NumChildren()
    1
    >>> list(vc.GetOnBits())
    [7, 9]


    to copy VectCollections, use the copy module:

    >>> bv1 = DataStructs.ExplicitBitVect(10)
    >>> bv1.SetBitsFromList((5,6))
    >>> vc.AddVect(3,bv1)
    >>> list(vc.GetOnBits())
    [5, 6, 7, 9]
    >>> vc2 = copy.copy(vc)
    >>> vc.DetachVectsNotMatchingBit(6)
    >>> list(vc.GetOnBits())
    [5, 6]
    >>> list(vc2.GetOnBits())
    [5, 6, 7, 9]

    The Uniquify() method can be used to remove duplicate vectors:

    >>> vc = VectCollection()
    >>> bv1 = DataStructs.ExplicitBitVect(10)
    >>> bv1.SetBitsFromList((7,9))
    >>> vc.AddVect(1,bv1)
    >>> vc.AddVect(2,bv1)
    >>> bv1 = DataStructs.ExplicitBitVect(10)
    >>> bv1.SetBitsFromList((2,3,5))
    >>> vc.AddVect(3,bv1)
    >>> vc.NumChildren()
    3
    >>> vc.Uniquify()
    >>> vc.NumChildren()
    2


    """

  def __init__(self):
    self.__vects = {}
    self.__orVect = None
    self.__numBits = -1
    self.__needReset = True

  def GetOrVect(self):
    if self.__needReset:
      self.Reset()
    return self.__orVect

  orVect = property(GetOrVect)

  def AddVect(self, idx, vect):
    self.__vects[idx] = vect
    self.__needReset = True

  def Reset(self):
    if not self.__needReset:
      return
    self.__orVect = None
    if not self.__vects:
      return
    ks = list(iter(self.__vects))
    self.__orVect = copy.copy(self.__vects[ks[0]])
    self.__numBits = self.__orVect.GetNumBits()
    for i in range(1, len(ks)):
      self.__orVect |= self.__vects[ks[i]]
    self.__needReset = False

  def NumChildren(self):
    return len(self.__vects.keys())

  def GetChildren(self):
    return tuple(self.__vects.items())

  def __getitem__(self, idx):
    if self.__needReset:
      self.Reset()
    return self.__orVect.GetBit(idx)

  GetBit = __getitem__

  def __len__(self):
    if self.__needReset:
      self.Reset()
    return self.__numBits

  GetNumBits = __len__

  def GetOnBits(self):
    if self.__needReset:
      self.Reset()
    return self.__orVect.GetOnBits()

  def DetachVectsNotMatchingBit(self, bit):
    items = list(self.__vects.items())
    for k, v in items:
      if not v.GetBit(bit):
        del (self.__vects[k])
        self.__needReset = True

  def DetachVectsMatchingBit(self, bit):
    items = list(self.__vects.items())
    for k, v in items:
      if v.GetBit(bit):
        del (self.__vects[k])
        self.__needReset = True

  def Uniquify(self, verbose=False):
    obls = {}
    for k, v in self.__vects.items():
      obls[k] = list(v.GetOnBits())

    keys = list(self.__vects.keys())
    nKeys = len(keys)
    keep = list(self.__vects.keys())
    for i in range(nKeys):
      k1 = keys[i]
      if k1 in keep:
        obl1 = obls[k1]
        idx = keys.index(k1)
        for j in range(idx + 1, nKeys):
          k2 = keys[j]
          if k2 in keep:
            obl2 = obls[k2]
            if obl1 == obl2:
              keep.remove(k2)

    self.__needsReset = True
    tmp = {}
    for k in keep:
      tmp[k] = self.__vects[k]
    if verbose:
      print('uniquify:', len(self.__vects), '->', len(tmp))
    self.__vects = tmp

  #
  # set up our support for pickling:
  #
  def __getstate__(self):
    pkl = struct.pack('<I', len(self.__vects))
    for k, v in self.__vects.items():
      pkl += struct.pack('<I', k)
      p = v.ToBinary()
      l = len(p)
      pkl += struct.pack('<I', l)
      pkl += struct.pack('%ds' % (l), p)
    return pkl

  def __setstate__(self, pkl):
    if isinstance(pkl, str):
      pkl = bytes(pkl, encoding='Latin1')

    self.__vects = {}
    self.__orVect = None
    self.__numBits = -1
    self.__needReset = True
    szI = struct.calcsize('I')
    offset = 0
    nToRead = struct.unpack('<I', pkl[offset:offset + szI])[0]
    offset += szI
    for _ in range(nToRead):
      k = struct.unpack('<I', pkl[offset:offset + szI])[0]
      offset += szI
      l = struct.unpack('<I', pkl[offset:offset + szI])[0]
      offset += szI
      sz = struct.calcsize('%ds' % l)
      bv = DataStructs.ExplicitBitVect(struct.unpack('%ds' % l, pkl[offset:offset + sz])[0])
      offset += sz
      self.AddVect(k, bv)


# ------------------------------------
#
#  doctest boilerplate
#
def _runDoctests(verbose=None):  # pragma: nocover
  import doctest
  import sys
  failed, _ = doctest.testmod(optionflags=doctest.ELLIPSIS, verbose=verbose)
  sys.exit(failed)


if __name__ == '__main__':  # pragma: nocover
  _runDoctests()



#
#  Copyright (c) 2016, Novartis Institutes for BioMedical Research Inc.
#  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#     * Neither the name of Novartis Institutes for BioMedical Research Inc.
#       nor the names of its contributors may be used to endorse or promote
#       products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Created by Nadine Schneider, July 2016

import itertools
from collections import Counter, defaultdict

import numpy as np

from rdkit import Chem
from rdkit.Chem import AllChem, rdqueries

from . import utils


class MoleculeDetails(object):

  __slots__ = [
    'detailFP', 'scaffoldFP', 'bitInfoDetailFP', 'bitInfoScaffoldFP', 'reactivity', 'bitReactivity',
    'molecule'
  ]

  def _atomDetailInvariant(self, mol):
    mol.UpdatePropertyCache(False)
    num_atoms = mol.GetNumAtoms()
    Chem.GetSSSR(mol)
    rinfo = mol.GetRingInfo()
    invariants = [0] * num_atoms
    for i, a in enumerate(mol.GetAtoms()):
      descriptors = []
      descriptors.append(a.GetAtomicNum())
      descriptors.append(a.GetTotalDegree())
      descriptors.append(a.GetTotalNumHs())
      descriptors.append(rinfo.IsAtomInRingOfSize(a.GetIdx(), 6))
      descriptors.append(rinfo.IsAtomInRingOfSize(a.GetIdx(), 5))
      descriptors.append(a.IsInRing())
      descriptors.append(a.GetIsAromatic())
      invariants[i] = hash(tuple(descriptors)) & 0xffffffff
    return invariants

  def _atomScaffoldInvariant(self, mol):
    num_atoms = mol.GetNumAtoms()
    invariants = [0] * num_atoms
    for i, a in enumerate(mol.GetAtoms()):
      descriptors = []
      descriptors.append(a.GetAtomicNum())
      invariants[i] = hash(tuple(descriptors)) & 0xffffffff
    return invariants

  def _createFP(self, mol, invariant, bitinfo, useBondTypes=True, radius=1):
    return AllChem.GetMorganFingerprint(mol=mol, radius=radius, invariants=invariant,
                                        useBondTypes=useBondTypes, bitInfo=bitinfo)

  def _isHeteroAtom(self, a):
    return a.GetAtomicNum() not in (6, 1)

  def _isSp3OrAromaticCarbon(self, a):
    if a.GetAtomicNum() != 6:
      return False
    if a.GetIsAromatic():
      return True
    for b in a.GetBonds():
      if b.GetBondTypeAsDouble() > 1.5:
        return False
    return True

  def _calcReactivityAtom(self, a):
    # exclude sp3 carbons or uncharged single heavy atoms such as water molecules
    if self._isSp3OrAromaticCarbon(a) or (len(a.GetNeighbors()) == 0 and a.GetFormalCharge() == 0):
      return 0
    # all other atoms have at least a reactivity of one
    reactivity = 1
    b = a.GetBonds()
    # if it is a heteroatom or has an H (we already know it's not SP3 or aromatic) increase the reactivity
    if self._isHeteroAtom(a) or a.GetTotalNumHs() > 0:
      reactivity += 1
    # slightly increase reactivity for atoms in aromatic rings compared to aliphatic rings
    if a.IsInRing():
      if a.GetIsAromatic():
        reactivity += 0.5
    # but prefer non-ring atoms
    else:
      reactivity += 1
    # increase reactivity of charged atoms
    if a.GetFormalCharge():
      reactivity += 2
    for bo in b:
      # look at the direct neighbors of the atom
      ni = bo.GetOtherAtom(a)
      # for non-single bonds increase the reactivity
      if bo.GetBondTypeAsDouble() > 1.5:
        reactivity += 1
        # if there are hydrogens attached, increase the reactivity
        if ni.GetTotalNumHs() > 0:
          reactivity += 1
      # if it is a bond to a hetero atom further increase the reactivity
      if self._isHeteroAtom(ni):
        reactivity += 1
        # bonds between nitrogens and oxygen or between oxygen and oxygen or between nitrogen and nitrogen are more reactive
        if a.GetAtomicNum() in (7, 8) and ni.GetAtomicNum() in (7, 8):
          reactivity += 2
        # if the neighbor is a Mg, Si, P, Pd, or Sn atom increase the reactivity
        elif ni.GetAtomicNum() in (12, 14, 15, 46, 50):
          reactivity += 1
    return reactivity

  def _calcReactivityMolecule(self, mol):
    reactivityAtoms = [self._calcReactivityAtom(a) for a in mol.GetAtoms()]
    return reactivityAtoms

  def __init__(self, molecule, verbose=0):
    self.molecule = molecule
    self.bitInfoDetailFP = {}
    self.detailFP = self._createFP(molecule, self._atomDetailInvariant(molecule),
                                   self.bitInfoDetailFP)
    self.bitInfoScaffoldFP = {}
    self.scaffoldFP = self._createFP(molecule, self._atomScaffoldInvariant(molecule),
                                     self.bitInfoScaffoldFP, useBondTypes=False)
    reactivityAtoms = self._calcReactivityMolecule(molecule)
    reactivity = sum(reactivityAtoms)
    if Chem.MolToSmiles(molecule) in frequentReagents:
      reactivity *= 0.8
    self.reactivity = reactivity


def _calcScore(reactantFP, productFP, bitInfoProd=None, output=False):
  if output:
    print("--- _calcScore ---")
  score = 0
  dFP = productFP - reactantFP
  numRBits = float(utils.getNumPositiveCounts(reactantFP))
  if output > 2:
    print("num RBits: ", numRBits)
  numPBits = float(utils.getNumPositiveCounts(productFP))
  if output > 2:
    print("num PBits: ", numPBits)
  numUnmappedPBits = float(utils.getNumPositiveCounts(dFP))
  if output > 2:
    print("num UnmappedPBits: ", numUnmappedPBits)
  numUnmappedRBits = float(utils.getNumNegativeCounts(dFP))
  if output > 2:
    print("num UnmappedRBits: ", numUnmappedRBits)

  numUnmappedPAtoms = -1
  bitsUnmappedPAtoms = -1
  if bitInfoProd is not None:
    numUnmappedPAtoms, bitsUnmappedPAtoms = utils.getNumPositiveBitCountsOfRadius0(dFP, bitInfoProd)
    if output > 2:
      print("num UnmappedPAtoms: ", numUnmappedPAtoms)
  ratioMappedPBits = 1 - (numUnmappedPBits / numPBits)
  ratioUnmappedRBits = numUnmappedRBits / numRBits
  score = max(ratioMappedPBits - ratioUnmappedRBits * ratioUnmappedRBits, 0)

  if output > 1:
    print("score: ", score, "(", ratioMappedPBits, ",", ratioUnmappedRBits * ratioUnmappedRBits,
          ",", ratioUnmappedRBits, ")")

  return [score, numUnmappedPBits, numUnmappedPAtoms, bitsUnmappedPAtoms]


# Set of frequent reagents derived from all patent reactions
frequentReagents = set([
  'CCN(CC)CC', '[Li+]', '[Na+]', 'O=C(O)CC(O)(CC(=O)O)C(=O)O', 'O=S(=O)(O)O', 'CN1CCCC1=O',
  'CCN(C(C)C)C(C)C', 'c1ccncc1', '[K]', 'CC(C)(C)O', 'CCO', 'Cc1ccc(S(=O)(=O)O)cc1',
  'ClC(Cl)(Cl)Cl', '[Na]', 'CC(C)(C)[O-]', 'O=C([O-])O', 'COCCOC', '[NH4+]', 'CC(C)OC(C)C',
  'O=C([O-])[O-]', 'CC(=O)OC(C)=O', 'O=C=O', '[Cl-]', 'c1ccc(P(c2ccccc2)c2ccccc2)cc1', '[H-]',
  'N#N', 'CN1CCOCC1', 'C1COCCO1', 'c1ccccc1', '[Cs+]', '[K+]', '[OH-]', 'CCCCCC', 'CCCCC',
  'CN(C)C=O', 'C[O-]', 'Cc1ccccc1', 'C1CCC2=NCCCN2CC1', 'CO', 'CCCCO', 'O=C(O)C(F)(F)F',
  'O=P([O-])([O-])[O-]', 'CCOC(C)=O', '[Mg+2]', 'C1CCCCC1', 'O', 'N', 'II', 'O=CO', 'CC(=O)N(C)C',
  'CC(=O)O', 'CCOCC', 'CC(C)O', 'C[Si](C)(C)Cl', 'Cc1ccccc1C', 'CC(C)=O', 'CS(=O)(=O)O',
  'CN(C)c1ccncc1', 'Cl', 'ClCCCl', 'O=S(Cl)Cl', 'ClC(Cl)Cl', '[Li]CCCC', '[Pd]', '[H][H]', '[Br-]',
  'CS(C)=O', 'COC(C)(C)C', 'O=S(=O)([O-])[O-]', 'CC(Cl)Cl', 'CC(=O)[O-]',
  'CCCC[N+](CCCC)(CCCC)CCCC', 'ClCCl', 'CC#N', 'C1CCOC1', 'CCCCCCC'
])


def _getBestCombination(rfps, pfps, output=False):

  if output:
    print("--- _getBestCombination ---")

  tests = []
  numReactants = len(rfps)
  # generate first all reactant combinations
  for i in range(1, numReactants + 1):
    for x in itertools.combinations(range(numReactants), i):
      temp = []
      for j in x:
        # don't include frequent reagents
        if not rfps[j][1]:
          numAtms = rfps[j][0].molecule.GetNumAtoms()
          # not test single ions
          if numAtms > 1:
            # store the number of reactant atoms for later
            temp.append((rfps[j][0].molecule.GetNumAtoms(), j))
        else:
          if output > 3:
            print("Frequent reagent found: ", j)
      if temp not in tests:
        tests.append(temp)
  # initialisation of the results
  maxScore = 0
  maxDetailScore = 0
  finalReacts = [[]]
  # get the product fingerprints
  productsDetailFP = utils.getSumFps([i.detailFP for i in pfps])
  productsScaffoldFP = utils.getSumFps([i.scaffoldFP for i in pfps])
  # get the number of atoms for the product
  numProductAtoms = 0
  for i in pfps:
    numProductAtoms += i.molecule.GetNumAtoms()
  # get the bitinfo for the product FP
  productsDetailFPBitInfo = {}
  productsScaffoldFPBitInfo = {}
  for i in pfps:
    productsDetailFPBitInfo.update(i.bitInfoDetailFP)
    productsScaffoldFPBitInfo.update(i.bitInfoScaffoldFP)
  # set some initial values
  numUnmappedPAtoms, bitsUnmappedPAtoms = utils.getNumPositiveBitCountsOfRadius0(
    productsScaffoldFP, productsScaffoldFPBitInfo)
  finalNumUnmappedProdAtoms = [[
    len(productsDetailFP.GetNonzeroElements()),
    len(productsScaffoldFP.GetNonzeroElements()), numUnmappedPAtoms, bitsUnmappedPAtoms
  ]]

  for test in tests:
    if len(test) < 1:
      continue
    # get the number of involved reactant atoms
    numReactantAtoms = np.array(test)[:, 0].sum()
    # ignore combinations including too many or too few atoms
    if numReactantAtoms > 5 * numProductAtoms or numReactantAtoms < numProductAtoms * 0.8:
      continue

    if output > 0:
      print("Combination: ", test)

    #build the combined reactant FPs
    reactantsDetailFP = utils.getSumFps([rfps[i[1]][0].detailFP for i in test])
    reactantsScaffoldFP = utils.getSumFps([rfps[i[1]][0].scaffoldFP for i in test])

    # get the scores for both FPs
    detailFPScore = _calcScore(reactantsDetailFP, productsDetailFP,
                               bitInfoProd=productsDetailFPBitInfo, output=output)
    scaffoldFPScore = _calcScore(reactantsScaffoldFP, productsScaffoldFP,
                                 bitInfoProd=productsScaffoldFPBitInfo, output=output)
    # final score
    score = detailFPScore[0] + scaffoldFPScore[0]

    if output > 0:
      print(">>>> score: ", score)
      print(">>>> scores (detail, scaffold): ", detailFPScore[0], scaffoldFPScore[0])
      print(">>>> num unmapped productFP bits: ", detailFPScore[1], scaffoldFPScore[1],
            detailFPScore[2], scaffoldFPScore[2])

    if score > maxScore:
      maxScore = score
      maxDetailScore = detailFPScore[0]
      del finalReacts[:]
      del finalNumUnmappedProdAtoms[:]
      # set the final reactants
      finalReacts.append([i[1] for i in test])
      # for tracking the mapping of the product atoms include the number of unmapped detailedFP bits, the number of unmapped
      # atoms based on the scaffold FP,  the number of unmapped scaffoldFP bits, and the unmapped scaffoldFP bits
      finalNumUnmappedProdAtoms.append(
        [detailFPScore[1], scaffoldFPScore[2], scaffoldFPScore[1], scaffoldFPScore[-1]])
      if output > 0:
        print(" >> maxScore: ", maxScore)
        print(" >> Final reactants: ", finalReacts)
      # test for almost perfect matchings (e.g. oxidations, reduction etc.)
      if scaffoldFPScore[0] > 0.9999 and detailFPScore[0] > 0.8:
        return finalReacts, finalNumUnmappedProdAtoms
      # test for number of mapped product atoms e.g. to capture deprotections earlier
      if len(finalNumUnmappedProdAtoms) > 0 and len(test) == 1:
        if finalNumUnmappedProdAtoms[0][1] == 0 and finalNumUnmappedProdAtoms[0][0] <= 3:
          return finalReacts, finalNumUnmappedProdAtoms
    # include alternative solutions
    elif abs(score - maxScore) < 0.0000001 and score > 0.0:
      finalReacts.append([i[1] for i in test])
      finalNumUnmappedProdAtoms.append(
        [detailFPScore[1], scaffoldFPScore[2], scaffoldFPScore[1], scaffoldFPScore[-1]])
      if output > 0:
        print(" >> Added alternative result")
        print(" >> Final reactants: ", finalReacts)

  return finalReacts, finalNumUnmappedProdAtoms


def _findMissingReactiveReactants(rfps, pfps, currentReactants, unmappedPAtoms, output=False):
  if output:
    print("--- _findMissingReactiveReactants ---")
  if not len(unmappedPAtoms):
    return currentReactants
  # if there are unmapped product bits find possible reactants for those
  else:
    finalReactants = []
    numReactants = len(rfps)
    # investigate all possible solutions of the scoring before
    for reacts, umPA in zip(currentReactants, unmappedPAtoms):
      # if there are unmapped product atoms find possible reactants for those
      finalReactants.append(reacts)
      if umPA[1] > 0:
        remainingReactants = set(range(numReactants)).difference(set(reacts))
        # sort the possible reactants by the reactivity
        remainingReactants = sorted(
          remainingReactants,
          key=lambda x: rfps[x].reactivity / float(rfps[x].molecule.GetNumAtoms()), reverse=True)
        missingPAtoms = []
        # get the missing atoms and counts
        for bit, c in umPA[-1]:
          for pbi in range(len(pfps)):
            if bit in pfps[pbi].bitInfoScaffoldFP:
              a = pfps[pbi].bitInfoScaffoldFP[bit][0]
              missingPAtoms.extend([pfps[pbi].molecule.GetAtomWithIdx(a[0]).GetAtomicNum()] * c)
        missingPAtoms = Counter(missingPAtoms)
        if output > 0:
          print(missingPAtoms)
        # build queries for the missing atoms
        queries = [(rdqueries.AtomNumEqualsQueryAtom(a), a) for a in missingPAtoms]
        maxFullfilledQueries = 0
        maxReactivity = -1
        addReactants = []
        # search for the most reactive reactants capturing all/most of the unmapped product atoms
        for r in remainingReactants:
          if output > 0:
            print(" >> Reactant", r, rfps[r].reactivity / float(rfps[r].molecule.GetNumAtoms()))
          countFullfilledQueries = 0
          for q, a in queries:
            if len(rfps[r].molecule.GetAtomsMatchingQuery(q)) >= missingPAtoms[a]:
              countFullfilledQueries += 1
          if output > 0:
            print(" Max reactivity", maxReactivity)
            print(" Max fulfilled queries", maxFullfilledQueries)
          if countFullfilledQueries > maxFullfilledQueries:
            maxFullfilledQueries = countFullfilledQueries
            maxReactivity = rfps[r].reactivity / float(rfps[r].molecule.GetNumAtoms())
            addReactants = [r]
          elif maxFullfilledQueries and countFullfilledQueries == maxFullfilledQueries and \
               rfps[r].reactivity/float(rfps[r].molecule.GetNumAtoms()) >= maxReactivity:
            maxFullfilledQueries = countFullfilledQueries
            addReactants.append(r)
          if output > 0:
            print(" Added reactants", addReactants)
        finalReactants[-1].extend(addReactants)
  if output > 0:
    print(" >> Final reactants", finalReactants)
  return finalReactants


def _detectObviousReagents(reactants, products):
  unchangedReacts = set()
  unchangedProds = set()
  for i, r in enumerate(reactants):
    for j, p in enumerate(products):
      if r == p:
        unchangedReacts.add(i)
        unchangedProds.add(j)
  return unchangedReacts, unchangedProds


def identifyReactants(reaction, output=False):
  rxn = AllChem.ChemicalReaction(reaction)
  AllChem.RemoveMappingNumbersFromReactions(rxn)
  if output:
    print("--- identifyReactants ---")
  reactants = rxn.GetReactants()
  products = rxn.GetProducts()
  ### Preprocessing
  uniqueReactants, reactantSmiles = utils.uniqueMolecules(reactants)
  uniqueProducts, productSmiles = utils.uniqueMolecules(products)
  # find molecules which do not change in the rxn
  unmodifiedReactants, unmodifiedProducts = _detectObviousReagents(reactantSmiles, productSmiles)
  if output:
    print("  >>> Found reagents in reactants:", unmodifiedReactants)
    print("  >>> Found reagents in products:", unmodifiedProducts)
  if len(products) == len(unmodifiedProducts):
    unmodifiedProducts = set()
  uniquePotentialReactants = [r for r in sorted(set(uniqueReactants.values()))]
  uniquePotentialProducts = [
    p for p in sorted(set(uniqueProducts.values())) if p not in unmodifiedProducts
  ]

  ### Find the most probable reactants
  # only generate moleculeDetail objects for unique, potential reactants and products
  rfps = [MoleculeDetails(reactants[r]) for r in uniquePotentialReactants]
  pfps = [MoleculeDetails(products[p]) for p in uniquePotentialProducts]

  rfpsPrep = [(MoleculeDetails(reactants[r]), reactantSmiles[r] in frequentReagents)
              for r in uniquePotentialReactants]

  reacts, unmappedProdAtoms = _getBestCombination(rfpsPrep, pfps, output=output)
  # no reactants where found try again including the frequent reagents
  if np.array(reacts).shape == (1, 0):
    rfpsPrep = [(MoleculeDetails(reactants[r]), 0) for r in uniquePotentialReactants]
    reacts, unmappedProdAtoms = _getBestCombination(rfpsPrep, pfps, output=output)

  ### Postprocessing
  # identify missing reactants
  reacts = _findMissingReactiveReactants(rfps, pfps, reacts, unmappedProdAtoms, output=output)
  finalreacts = []
  for i in reacts:
    temp = [uniquePotentialReactants[j] for j in i]
    finalreacts.append(set(temp))

  return finalreacts, unmodifiedReactants, unmodifiedProducts


# reassign the reaction roles of a reaction
def reassignRXNRoles(rxn):
  utils.transferAgentsToReactants(rxn)
  reacts, rAgents, pAgents = identifyReactants(rxn)
  if len(reacts) < 1:
    return None
  new_rxn = AllChem.ChemicalReaction()
  for i in range(rxn.GetNumProductTemplates()):
    new_rxn.AddProductTemplate(rxn.GetProductTemplate(i))
  for i in range(rxn.GetNumReactantTemplates()):
    if i in reacts[0]:
      new_rxn.AddReactantTemplate(rxn.GetReactantTemplate(i))
    else:
      new_rxn.AddAgentTemplate(rxn.GetReactantTemplate(i))
  return new_rxn


# clean-up the reaction smiles
def reassignReactionRoles(smi):
  rxn = AllChem.ReactionFromSmarts(smi, useSmiles=True)
  new_rxn = reassignRXNRoles(rxn)
  if new_rxn is None:
    return ''
  smi_new = AllChem.ReactionToSmiles(new_rxn)
  return smi_new



import argparse
import operator
import sys
from collections import Counter, defaultdict, namedtuple

import numpy as np
import pandas as pd

from rdkit import Chem
from rdkit.Chem import FilterCatalog, RDConfig, rdMolDescriptors

FilterMatch = namedtuple(
  'FilterMatch',
  ('SubstructureMatches', 'Min_N_O_filter', 'Frac_N_O', 'Covalent', 'SpecialMol', 'SeverityScore'))


# Build the filter catalog using the RDKit filterCatalog module
def buildFilterCatalog():

  inhousefilter = pd.read_csv(
    f'{RDConfig.RDContribDir}/NIBRSubstructureFilters/SubstructureFilter_HitTriaging_wPubChemExamples.csv'
  )
  inhouseFiltersCat = FilterCatalog.FilterCatalog()
  for i in range(inhousefilter.shape[0]):
    mincount = 1
    if inhousefilter['MIN_COUNT'][i] != 0:
      mincount = int(inhousefilter['MIN_COUNT'][i])
    pname = inhousefilter['PATTERN_NAME'][i]
    sname = inhousefilter['SET_NAME'][i]
    pname_final = '{0}_min({1})__{2}__{3}__{4}'.format(pname, mincount,
                                                       inhousefilter['SEVERITY_SCORE'][i],
                                                       inhousefilter['COVALENT'][i],
                                                       inhousefilter['SPECIAL_MOL'][i])
    fil = FilterCatalog.SmartsMatcher(pname_final, inhousefilter['SMARTS'][i], mincount)
    inhouseFiltersCat.AddEntry(FilterCatalog.FilterCatalogEntry(pname_final, fil))
    inhouseFiltersCat.GetEntry(i).SetProp('Scope', sname)
  return inhouseFiltersCat


# Assign substructure filters and fraction of Nitrogen and Oxygen atoms
def assignFilters(data, nameSmilesColumn='smiles'):

  results = []

  inhouseFiltersCat = buildFilterCatalog()

  NO_filter = '[#7,#8]'
  sma = Chem.MolFromSmarts(NO_filter, mergeHs=True)

  for smi in data[nameSmilesColumn]:
    qc, NO_filter, fracNO, co, sc, sm = [np.nan] * 6

# The following files require numpy and were explicitely checked for their compatibility.

    try:
      mol = Chem.MolFromSmiles(smi)

      # fraction of N and O atoms
      numHeavyAtoms = mol.GetNumHeavyAtoms()
      numNO = len(mol.GetSubstructMatches(Chem.MolFromSmarts('[#7,#8]')))
      fracNO = float(numNO) / numHeavyAtoms

      # all substructure filters
      entries = inhouseFiltersCat.GetMatches(mol)
      if len(list(entries)):
        # initialize empty lists
        fs, sev, cov, spm = ([] for _ in range(4))
        # get the matches
        for entry in entries:
          pname = entry.GetDescription()
          n, s, c, m = pname.split('__')
          fs.append(entry.GetProp("Scope") + '_' + n)
          sev.append(int(s))
          cov.append(int(c))
          spm.append(int(m))
        # concatenate all matching filters
        qc = ' | '.join(fs)
        # assign overall severity
        if sev.count(2):
          sc = 10
        else:
          sc = sum(sev)
        # get number of covalent flags and special molecule flags
        co = sum(cov)
        sm = sum(spm)
      # if non of the filters matches
      else:
        qc = 'no match'
        sc = 0
        co = 0
        sm = 0

      # special NO filter
      if not mol.HasSubstructMatch(sma):
        NO_filter = 'no_oxygen_or_nitrogen'
      else:
        NO_filter = 'no match'
    except Exception:
      print("Failed on compound {0}\n".format(smi))
      pass
    results.append(FilterMatch(qc, NO_filter, fracNO, co, sm, sc))
  return results


if __name__ == "__main__":
  parser = argparse.ArgumentParser()
  parser.add_argument('--data', type=str, required=True,
                      help='Please specify the path to your data file. Required format: csv')
  parser.add_argument('--smilesColumn', type=str, required=True,
                      help='Please specify the name of your SMILES column.')
  parser.add_argument('--result', type=str, required=True,
                      help='Please specify the name of your result file.')
  parser.add_argument('--verbose', type=bool, default=1, help='Generate output? Default: False')
  args = parser.parse_args()

  if args.verbose:
    print('---> Reading data')
  datafile = args.data
  try:
    data = pd.read_csv(datafile)
  except Exception:
    if args.verbose:
      print('Data could not be read. Please check your file.')
    sys.exit()

  smiCol = args.smilesColumn

  if args.verbose:
    print('---> Apply filters to data')
  try:
    results = assignFilters(data, nameSmilesColumn=smiCol)
  except Exception:
    if args.verbose:
      print('Smiles column does not exist. Please check.')
    sys.exit()

  df_tmp = pd.DataFrame.from_records(results, columns=FilterMatch._fields)

  data = data.merge(df_tmp, how='left', left_index=True, right_index=True)
  data.to_csv(args.result, index=False)

  if args.verbose:
    print('---> Done')



#
# Copyright (C) 2003 Rational Discovery LLC
#    All Rights Reserved
#

availableCanvases = {
  'pdf': ('PDF.pidPDF', 'PDFCanvas', 'PDF'),
  'ps': ('PS.pidPS', 'PSCanvas', 'PS'),
  'svg': ('SVG.pidSVG', 'SVGCanvas', 'SVG'),
  'jpg': ('PIL.pidPIL', 'PILCanvas', 'JPEG'),
  'png': ('PIL.pidPIL', 'PILCanvas', 'PNG')
}



# Copyright (C) 2007-2024 Greg Landrum and other RDKit contributors
#   All Rights Reserved
#
""" Implementation of the clustering algorithm published in:
  Butina JCICS 39 747-750 (1999)

"""
import numpy as np

def EuclideanDist(pi, pj):
  """Calculate the Euclidean distance between two points."""
  pi, pj = np.asarray(pi), np.asarray(pj)
  return np.sqrt(np.sum((pi - pj) ** 2))

def compute_distance_matrix(data, n_pts, dist_func):
  """Compute the distance matrix for the given data."""
  dist_matrix = np.zeros((n_pts, n_pts))
  for i in range(n_pts):
    for j in range(i):
      dist_matrix[i, j] = dist_matrix[j, i] = dist_func(data[i], data[j])
  return dist_matrix

def ClusterData(data, nPts, distThresh, isDistData=False, distFunc=EuclideanDist, reordering=False):
  """  clusters the data points passed in and returns the list of clusters

    **Arguments**

      - data: a list, tuple, or numpy array of items with the input data
        (see discussion of _isDistData_ argument for the exception)

      - nPts: the number of points to be used

      - distThresh: elements within this range of each other are considered
        to be neighbors

      - isDistData: set this toggle when the data passed in is a
          distance matrix.  The distance matrix should be stored 
          in one of two formats: as an nxn NumPy array, or as a 
          symmetrically stored list or 1D array generated using a 
          similar process to the example below:

            dists = []
            for i in range(nPts):
              for j in range(i):
                dists.append( distfunc(i,j) )

      - distFunc: a function to calculate distances between points.
           Receives 2 points as arguments, should return a float

      - reordering: if this toggle is set, the number of neighbors is updated
           for the unassigned molecules after a new cluster is created such
           that always the molecule with the largest number of unassigned
           neighbors is selected as the next cluster center.

    **Returns**

      - a tuple of tuples containing information about the clusters:
         ( (cluster1_elem1, cluster1_elem2, ...),
           (cluster2_elem1, cluster2_elem2, ...),
           ...
         )
         The first element for each cluster is its centroid.

  """
  if isDistData:
    # Check if data is a supported type
    if not isinstance(data, (list, tuple, np.ndarray)):
      raise TypeError(f"Unsupported type for data, {type(data)}")
    
    # Check if data is a 1D array or list
    if isinstance(data, (list, tuple)) or (isinstance(data, np.ndarray) and data.ndim == 1):
      # Check if data length matches the required number of points
      if len(data) != (nPts * (nPts - 1)) // 2:
        raise ValueError("Mismatched input data dimension and nPts")
        
      # Create a distance matrix from the 1D data
      dist_matrix = np.zeros((nPts, nPts))
      idx = np.tril_indices(nPts, -1)
      dist_matrix[idx] = data
      dist_matrix += dist_matrix.T
    else:
      # Check if data is a matrix of the correct shape and use it as distance matrix
      if data.shape != (nPts, nPts):
          raise ValueError(f"Input data with shape {data.shape} is not a matrix of the required shape {(nPts, nPts)}")
      dist_matrix = data
  else:
    # Compute distance matrix from the data points
    dist_matrix = compute_distance_matrix(data, nPts, distFunc)

  # Initialize neighbor lists
  neighbor_lists = [np.where(dist_matrix[i] <= distThresh)[0].tolist() for i in range(nPts)]

  # Sort points by the number of neighbors in descending order
  sorted_indices = [(len(neighbors), idx) for idx, neighbors in enumerate(neighbor_lists)]
  sorted_indices.sort(reverse=True)

  # Initialize clusters and a seen array to keep track of processed points
  clusters = []
  seen = np.zeros(nPts, dtype=bool)

  # Process all candidate clusters that have at least two members
  while sorted_indices and sorted_indices[0][0] > 1:
    _, idx = sorted_indices.pop(0)
    if seen[idx]:
      continue

    # Create a new cluster and mark points as seen
    cluster = [idx]
    seen[idx] = True
    for neighbor in neighbor_lists[idx]:
      if not seen[neighbor]:
        cluster.append(neighbor)
        seen[neighbor] = True
    
    clusters.append(tuple(cluster))

    # Update the number of neighbors:
    # remove all members of the new cluster from the list of
    # neighbors and reorder the sorted_indices
    if reordering:
      # Get the set of unassigned and affected molecules, i.e. all unseen molecules
      # which have at least one of the members of the new cluster
      # as a neighbor
      affected = set(neighbor for point in cluster for neighbor in neighbor_lists[point] if not seen[neighbor])

      # Loop over all remaining molecules in sorted_indices but only
      # consider unassigned and affected compounds
      for ii, element in enumerate(sorted_indices):
        affected_point = element[1]
        if affected_point in affected:
          # Update the number of neighbors
          new_neighbors = [nbr for nbr in neighbor_lists[affected_point] if not seen[nbr]]
          neighbor_lists[affected_point] = new_neighbors
          sorted_indices[ii] = (len(new_neighbors), affected_point)
      # Reorder the list
      sorted_indices.sort(reverse=True)
  
  # Process any remaining single-point clusters
  while sorted_indices:
    _, idx = sorted_indices.pop(0)
    if seen[idx]:
      continue
    clusters.append(tuple([idx]))
  return tuple(clusters)



#
# Copyright (C) 2006-2021 Greg Landrum
#  All Rights Reserved
#
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.
#
import os
import warnings
from collections import namedtuple
from importlib.util import find_spec
from io import BytesIO
import itertools

import numpy
from rdkit import Chem
from rdkit import RDConfig
from rdkit import rdBase
from rdkit.Chem import rdDepictor
from rdkit.Chem.Draw import rdMolDraw2D
from rdkit.Chem.Draw.rdMolDraw2D import *


def _sip_available():
  try:
    from rdkit.Chem.Draw.rdMolDraw2DQt import rdkitQtVersion
  except ImportError:
    return False
  pyqt_pkg = f'PyQt{rdkitQtVersion[0]}'
  if find_spec(pyqt_pkg) and find_spec(f'{pyqt_pkg}.sip'):
    return True
  elif find_spec('sip'):
    return True
  return False


if _sip_available():

  def MolDraw2DFromQPainter(qpainter, width=-1, height=-1, panelWidth=-1, panelHeight=-1):
    from rdkit.Chem.Draw import rdMolDraw2DQt
    if rdMolDraw2DQt.rdkitQtVersion.startswith('6'):
      from PyQt6.QtGui import QPainter
    else:
      from PyQt5.Qt import QPainter
    try:
      # Prefer the PyQt-bundled sip
      if rdMolDraw2DQt.rdkitQtVersion.startswith('6'):
        from PyQt6 import sip
      else:
        from PyQt5 import sip
    except ImportError:
      # No bundled sip, try the standalone package
      import sip

    if not isinstance(qpainter, QPainter):
      raise ValueError("argument must be a QPainter instance")
    if width <= 0:
      width = qpainter.viewport().width()
    if height <= 0:
      height = qpainter.viewport().height()
    ptr = sip.unwrapinstance(qpainter)
    d2d = rdMolDraw2DQt.MolDraw2DFromQPainter_(width, height, ptr, panelWidth, panelWidth)
    # tie the lifetime of the QPainter to this MolDraw2D object
    d2d._qptr = qpainter
    return d2d


def MolToImage(mol, size=(300, 300), kekulize=True, wedgeBonds=True, fitImage=False, options=None,
               **kwargs):
  """Returns a PIL image containing a drawing of the molecule.

      ARGUMENTS:

        - kekulize: run kekulization routine on input `mol` (default True)

        - size: final image size, in pixel (default (300,300))

        - wedgeBonds: draw wedge (stereo) bonds (default True)

        - highlightAtoms: list of atoms to highlight (default [])

        - highlightBonds: list of bonds to highlight (default [])

        - highlightColor: RGB color as tuple (default [1, 0, 0])

        - legend: optional text drawn as legend (default ''). Position and style
          are controlled by options (MolDrawOptions): legendPosition (Bottom, Top,
          Left, Right), legendVerticalText (for Left/Right), legendFontSize,
          legendFraction. Pass as options=opts or drawOptions=opts.

        - options: rdMolDraw2D.MolDrawOptions instance for drawing options
          (e.g. legendPosition, legendVerticalText). Single-molecule only.

      NOTE:

            use 'matplotlib.colors.to_rgb()' to convert string and
            HTML color codes into the RGB tuple representation, eg.

              from matplotlib.colors import ColorConverter
              img = Draw.MolToImage(m, highlightAtoms=[1,2], highlightColor=ColorConverter().to_rgb('aqua'))
              img.save("molecule.png")

      RETURNS:

        a PIL Image object
  """
  if not mol:
    raise ValueError('Null molecule provided')
  if not hasattr(rdMolDraw2D, 'MolDraw2DCairo'):
    raise RuntimeError("MolToImage requires that the RDKit be built with Cairo support")
  return _moltoimg(mol, size, kwargs.get('highlightAtoms', []), kwargs.get('legend', ''),
                   highlightBonds=kwargs.get('highlightBonds',
                                             []), drawOptions=options, kekulize=kekulize,
                   wedgeBonds=wedgeBonds, highlightColor=kwargs.get('highlightColor', None))


def MolToFile(mol, filename, size=(300, 300), kekulize=True, wedgeBonds=True, imageType=None,
              fitImage=False, options=None, **kwargs):
  """ Generates a drawing of a molecule and writes it to a file
  """
  # original contribution from Uwe Hoffmann
  if not filename:
    raise ValueError('no filename provided')
  if not mol:
    raise ValueError('Null molecule provided')

  if imageType is None:
    imageType = os.path.splitext(filename)[1][1:]

  if imageType not in ('svg', 'png'):
    raise ValueError('only PNG and SVG image types are supported')

  if imageType == 'png':
    drawfn = _moltoimg
    mode = 'b'
  elif imageType == 'svg':
    drawfn = _moltoSVG
    mode = 't'
  else:
    raise ValueError("unsupported output format")
  data = drawfn(mol, size, kwargs.get('highlightAtoms', []), kwargs.get('legend', ''),
                highlightBonds=kwargs.get('highlightBonds', []), drawOptions=options,
                kekulize=kekulize, wedgeBonds=wedgeBonds, returnPNG=True)
  with open(filename, 'w+' + mode) as outf:
    outf.write(data)
    outf.close()


def MolToSVG(mol, size=(300, 300), kekulize=True, wedgeBonds=True, drawOptions=None, **kwargs):
  """Returns an SVG string containing a drawing of the molecule.

  Supports the same arguments as MolToImage for highlights and legend.
  For legend position (Top/Left/Right/Bottom) and vertical text on side legends,
  pass an rdMolDraw2D.MolDrawOptions instance with legendPosition and
  legendVerticalText set. Legend position applies to single-molecule drawing;
  for grid drawing (MolsToGridImage), pass the same drawOptions and the position
  applies to each cell's legend.
  """
  if not mol:
    raise ValueError('Null molecule provided')
  return _moltoSVG(mol, size, kwargs.get('highlightAtoms', []), kwargs.get('legend', ''), kekulize,
                   drawOptions=drawOptions, highlightBonds=kwargs.get('highlightBonds', []))


def ShowMol(mol, size=(300, 300), kekulize=True, wedgeBonds=True, title='RDKit Molecule',
            stayInFront=True, **kwargs):
  """ Generates a picture of a molecule and displays it in a Tkinter window
  """
  import tkinter

  from PIL import ImageTk

  img = MolToImage(mol, size, kekulize, wedgeBonds, **kwargs)

  tkRoot = tkinter.Tk()
  tkRoot.title(title)
  tkPI = ImageTk.PhotoImage(img)
  tkLabel = tkinter.Label(tkRoot, image=tkPI)
  tkLabel.place(x=0, y=0, width=img.size[0], height=img.size[1])
  tkRoot.geometry('%dx%d' % (img.size))
  tkRoot.lift()
  if stayInFront:
    tkRoot.attributes('-topmost', True)
  tkRoot.mainloop()


def _bivariate_normal(X, Y, sigmax=1.0, sigmay=1.0, mux=0.0, muy=0.0, sigmaxy=0.0):
  """

    This is the implementation from matplotlib:
    https://github.com/matplotlib/matplotlib/blob/81e8154dbba54ac1607b21b22984cabf7a6598fa/lib/matplotlib/mlab.py#L1866
    it was deprecated in v2.2 of matplotlib, so we are including it here.


    Bivariate Gaussian distribution for equal shape *X*, *Y*.
    See `bivariate normal
    <http://mathworld.wolfram.com/BivariateNormalDistribution.html>`_
    at mathworld.
    """
  Xmu = X - mux
  Ymu = Y - muy

  rho = sigmaxy / (sigmax * sigmay)
  z = Xmu**2 / sigmax**2 + Ymu**2 / sigmay**2 - 2 * rho * Xmu * Ymu / (sigmax * sigmay)
  denom = 2 * numpy.pi * sigmax * sigmay * numpy.sqrt(1 - rho**2)
  return numpy.exp(-z / (2 * (1 - rho**2))) / denom


def calcAtomGaussians(mol, a=0.03, step=0.02, weights=None):
  """
  """
  x = numpy.arange(0, 1, step)
  y = numpy.arange(0, 1, step)
  X, Y = numpy.meshgrid(x, y)
  if weights is None:
    weights = [1.] * mol.GetNumAtoms()
  Z = _bivariate_normal(X, Y, a, a, mol._atomPs[0][0], mol._atomPs[0][1]) * weights[0]
  for i in range(1, mol.GetNumAtoms()):
    Zp = _bivariate_normal(X, Y, a, a, mol._atomPs[i][0], mol._atomPs[i][1])
    Z += Zp * weights[i]
  return X, Y, Z


def MolsToImage(mols, subImgSize=(200, 200), legends=None, **kwargs):
  """
  """
  from PIL import Image
  if legends is None:
    legends = [None] * len(mols)
  res = Image.new("RGBA", (subImgSize[0] * len(mols), subImgSize[1]))
  for i, mol in enumerate(mols):
    res.paste(MolToImage(mol, subImgSize, legend=legends[i], **kwargs), (i * subImgSize[0], 0))
  return res


def _drawerToImage(d2d):
  from PIL import Image
  sio = BytesIO(d2d.GetDrawingText())
  return Image.open(sio)


def shouldKekulize(mol, kekulize):
  if kekulize:
    return not any(bond.GetIsAromatic() and bond.HasQuery() for bond in mol.GetBonds())
  return kekulize


def _prepareMol(mol, drawOptions, **kwargs):
  if mol.NeedsUpdatePropertyCache():
    mol.UpdatePropertyCache(False)

  kekulize = shouldKekulize(mol, kwargs.get('kekulize', True))
  wedge = kwargs.get('wedgeBonds', True)

  if not drawOptions or drawOptions.prepareMolsBeforeDrawing:
    try:
      with rdBase.BlockLogs():
        mol = rdMolDraw2D.PrepareMolForDrawing(mol, kekulize=kekulize, wedgeBonds=wedge)
    except ValueError:  # <- can happen on a kekulization failure
      mol = rdMolDraw2D.PrepareMolForDrawing(mol, kekulize=False, wedgeBonds=wedge)
  return mol


def _moltoimg(mol, sz, highlights, legend, returnPNG=False, drawOptions=None, **kwargs):
  mol = _prepareMol(mol, drawOptions=drawOptions, **kwargs)
  d2d = rdMolDraw2D.MolDraw2DCairo(sz[0], sz[1])
  if drawOptions is not None:
    d2d.SetDrawOptions(drawOptions)
  if 'highlightColor' in kwargs and kwargs['highlightColor']:
    d2d.drawOptions().setHighlightColour(kwargs['highlightColor'])
  # we already prepared the molecule:
  d2d.drawOptions().prepareMolsBeforeDrawing = False
  bondHighlights = kwargs.get('highlightBonds', None)
  if highlights:
    if isinstance(highlights[0], (list, tuple)):
      # multiple highlights
      highlights = list(itertools.chain.from_iterable(highlights))
  if bondHighlights is not None:
    d2d.DrawMolecule(mol, legend=legend or "", highlightAtoms=highlights or [],
                     highlightBonds=bondHighlights)
  else:
    d2d.DrawMolecule(mol, legend=legend or "", highlightAtoms=highlights or [])
  d2d.FinishDrawing()
  if returnPNG:
    img = d2d.GetDrawingText()
  else:
    img = _drawerToImage(d2d)
  return img


def _moltoSVG(mol, sz, highlights, legend, kekulize, drawOptions=None, **kwargs):
  if mol.NeedsUpdatePropertyCache():
    mol.UpdatePropertyCache(False)

  kekulize = shouldKekulize(mol, kekulize)

  if not drawOptions or drawOptions.prepareMolsBeforeDrawing:
    try:
      with rdBase.BlockLogs():
        mol = rdMolDraw2D.PrepareMolForDrawing(mol, kekulize=kekulize)
    except ValueError:  # <- can happen on a kekulization failure
      mol = rdMolDraw2D.PrepareMolForDrawing(mol, kekulize=False)
  d2d = rdMolDraw2D.MolDraw2DSVG(sz[0], sz[1])
  if drawOptions is not None:
    d2d.SetDrawOptions(drawOptions)
  # we already prepared the molecule:
  d2d.drawOptions().prepareMolsBeforeDrawing = False
  bondHighlights = kwargs.get('highlightBonds', None)
  if bondHighlights is not None:
    d2d.DrawMolecule(mol, legend=legend or "", highlightAtoms=highlights or [],
                     highlightBonds=bondHighlights)
  else:
    d2d.DrawMolecule(mol, legend=legend or "", highlightAtoms=highlights or [])
  d2d.FinishDrawing()
  svg = d2d.GetDrawingText()
  return svg


def _MolsToGridImage(mols, molsPerRow=3, subImgSize=(200, 200), legends=None,
                     highlightAtomLists=None, highlightBondLists=None, drawOptions=None,
                     returnPNG=False, **kwargs):
  """ returns a PIL Image of the grid
  """
  if legends is None:
    legends = [''] * len(mols)

  nRows = len(mols) // molsPerRow
  if len(mols) % molsPerRow:
    nRows += 1

  if not hasattr(rdMolDraw2D, 'MolDraw2DCairo'):
    raise RuntimeError("MolsToGridImage requires that the RDKit be built with Cairo support")
  fullSize = (molsPerRow * subImgSize[0], nRows * subImgSize[1])
  d2d = rdMolDraw2D.MolDraw2DCairo(fullSize[0], fullSize[1], subImgSize[0], subImgSize[1])
  if drawOptions is not None:
    d2d.SetDrawOptions(drawOptions)
  else:
    dops = d2d.drawOptions()
    for k, v in list(kwargs.items()):
      if hasattr(dops, k):
        setattr(dops, k, v)
        del kwargs[k]
  d2d.DrawMolecules(list(mols), legends=legends or None, highlightAtoms=highlightAtomLists,
                    highlightBonds=highlightBondLists, **kwargs)
  d2d.FinishDrawing()
  if not returnPNG:
    res = _drawerToImage(d2d)
  else:
    res = d2d.GetDrawingText()

  return res


def _padList(inputList, lengthShouldBe, padWith=""):
  length = len(inputList)
  paddingCount = lengthShouldBe - length
  paddedList = inputList + [padWith] * paddingCount
  return paddedList


def _padMatrix(inputMatrix, rowLength, padWith=""):
  paddedMatrix = [_padList(row, rowLength, padWith) for row in inputMatrix]
  return paddedMatrix


def _flattenTwoDList(twoDList):
  return [item for sublist in twoDList for item in sublist]


def _MolsNestedToLinear(molsMatrix, legendsMatrix=None, highlightAtomListsMatrix=None,
                        highlightBondListsMatrix=None):
  """Converts a nested data structure (where each data substructure represents a row in mol grid image)
  to a linear one, padding rows as needed so all rows are the length of the longest row
  """
  # Check that other matrices (if provided) are same length,
  #   and each element (sub-iterable) is the same length, as molsMatrix

  nMolsRows = len(molsMatrix)

  if legendsMatrix is not None:
    nLegendsRows = len(legendsMatrix)
    if nLegendsRows != nMolsRows:
      err = f"If legendsMatrix is provided it must be the same length (have the same number "
      err += f"of sub-iterables) as molsMatrix, {nMolsRows}; its length is {nLegendsRows}."
      raise ValueError(err)
    for rowIndex, row in enumerate(legendsMatrix):
      if len(row) != len(molsMatrix[rowIndex]):
        err = f"If legendsMatrix is provided each of its sub-iterables must be the same length "
        err += f"as the corresponding sub-iterable of molsMatrix. For sub-iterable of index "
        err += f"{rowIndex}, its length in molsMatrix is {len(molsMatrix[rowIndex])} "
        err += f"while its length in legendsMatrix is {len(row)}."
        raise ValueError(err)

  if highlightAtomListsMatrix is not None:
    nHighlightAtomListsRows = len(highlightAtomListsMatrix)
    if nHighlightAtomListsRows != nMolsRows:
      err = f"If highlightAtomListsMatrix is provided it must be the same length (have the same number "
      err += f"of sub-iterables) as molsMatrix, {nMolsRows}; its length is {nHighlightAtomListsRows}."
      raise ValueError(err)
    for rowIndex, row in enumerate(highlightAtomListsMatrix):
      if len(row) != len(molsMatrix[rowIndex]):
        err = f"If highlightAtomListsMatrix is provided each of its sub-iterables must be the same length "
        err += f"as the corresponding sub-iterable of molsMatrix. For sub-iterable of index "
        err += f"{rowIndex}, its length in molsMatrix is {len(molsMatrix[rowIndex])} "
        err += f"while its length in highlightAtomListsMatrix is {len(row)}."
        raise ValueError(err)

  if highlightBondListsMatrix is not None:
    nHighlightBondListsRows = len(highlightBondListsMatrix)
    if nHighlightBondListsRows != nMolsRows:
      err = f"If highlightBondListsMatrix is provided it must be the same length (have the same number "
      err += f"of sub-iterables) as molsMatrix, {nMolsRows}; its length is {nHighlightBondListsRows}."
      raise ValueError(err)
    for rowIndex, row in enumerate(highlightBondListsMatrix):
      if len(row) != len(molsMatrix[rowIndex]):
        err = f"If highlightBondListsMatrix is provided each of its sub-iterables must be the same length "
        err += f"as the corresponding sub-iterable of molsMatrix. For sub-iterable of index "
        err += f"{rowIndex}, its length in molsMatrix is {len(molsMatrix[rowIndex])} "
        err += f"while its length in highlightBondListsMatrix is {len(row)}."
        raise ValueError(err)

  molsPerRow = max(len(row) for row in molsMatrix)

  # Pad matrices so they're rectangular (same length for each sublist),
  #   then convert to 1D lists
  # Pad using None for molecule for empty cells
  molsMatrixPadded = _padMatrix(molsMatrix, molsPerRow, None)
  mols = _flattenTwoDList(molsMatrixPadded)

  if legendsMatrix is not None:
    legendsMatrixPadded = _padMatrix(legendsMatrix, molsPerRow, "")
    legends = _flattenTwoDList(legendsMatrixPadded)
  else:
    legends = None

  if highlightAtomListsMatrix is not None:
    highlightAtomListsPadded = _padMatrix(highlightAtomListsMatrix, molsPerRow, [])
    highlightAtomLists = _flattenTwoDList(highlightAtomListsPadded)
  else:
    highlightAtomLists = None

  if highlightBondListsMatrix is not None:
    highlightBondListsPadded = _padMatrix(highlightBondListsMatrix, molsPerRow, [])
    highlightBondLists = _flattenTwoDList(highlightBondListsPadded)
  else:
    highlightBondLists = None

  return mols, molsPerRow, legends, highlightAtomLists, highlightBondLists


def MolsMatrixToGridImage(molsMatrix, subImgSize=(200, 200), legendsMatrix=None,
                          highlightAtomListsMatrix=None, highlightBondListsMatrix=None,
                          useSVG=False, returnPNG=False, **kwargs):
  r"""Creates a mol grid image from a nested data structure (where each data substructure represents a row),
  padding rows as needed so all rows are the length of the longest row

      ARGUMENTS:

        - molsMatrix: A two-deep nested data structure of RDKit molecules to draw,
          iterable of iterables (for example list of lists) of RDKit molecules

        - subImgSize: The size of a cell in the drawing; passed through to MolsToGridImage (default (200, 200))

        - legendsMatrix: A two-deep nested data structure of strings to label molecules with,
          iterable of iterables (for example list of lists) of strings (default None)

        - highlightAtomListsMatrix: A three-deep nested data structure of integers of atoms to highlight,
          iterable of iterables (for example list of lists) of integers (default None)

        - highlightBondListsMatrix: A three-deep nested data structure of integers of bonds to highlight,
          iterable of iterables (for example list of lists) of integers (default None)

        - useSVG: Whether to return an SVG (if true) or PNG (if false);
          passed through to MolsToGridImage (default false)

        - returnPNG: Whether to return PNG data (if true) or a PIL object for a PNG image file (if false);
          has no effect if useSVG is true; passed through to MolsToGridImage (default false)

        - kwargs: Any other keyword arguments are passed to MolsToGridImage

      NOTES:

            To include a blank cell in the middle of a row, supply None for that entry in molsMatrix.
            You do not need to do that for empty cells at the end of a row; 
            this function will automatically pad rows so that all rows are the same length.
            
            This function is useful when each row has some meaning,
            for example the generation in a mass spectrometry fragmentation tree--refer to 
            example at https://en.wikipedia.org/wiki/Fragmentation_(mass_spectrometry).
            If you want to display a set molecules where each row does not have any specific meaning,
            use MolsToGridImage instead.

            This function nests data structures one additional level beyond the analogous function MolsToGridImage
            (in which the molecules and legends are non-nested lists, 
            and the highlight parameters are two-deep nested lists) 

      RETURNS:

        A grid of molecular images in one of these formats:
        
        - useSVG=False and returnPNG=False (default): A PIL object for a PNG image file

        - useSVG=False and returnPNG=True: PNG data

        - useSVG=True: An SVG string

      EXAMPLES:

        from rdkit import Chem
        from rdkit.Chem.Draw import MolsMatrixToGridImage, rdMolDraw2D
        FCl = Chem.MolFromSmiles("FCl")
        molsMatrix = [[FCl, FCl], [FCl, None, FCl]]

        # Minimal example: Only molsMatrix is supplied,
        # result will be a drawing containing (where each row contains molecules):
        # F-Cl    F-Cl
        # F-Cl            F-Cl
        img = MolsMatrixToGridImage(molsMatrix)
        img.save("MolsMatrixToGridImageMinimal.png")
        # img is a PIL object for a PNG image file like:
        # <PIL.PngImagePlugin.PngImageFile image mode=RGB size=600x200 at 0x1648CC390>
        # Drawing will be saved as PNG file MolsMatrixToGridImageMinimal.png

        # Exhaustive example: All parameters are supplied,
        # result will be a drawing containing (where each row of molecules is followed by a row of legends):
        # 1 F-Cl 0              1 F-Cl 0
        # no highlighting       bond highlighted         
        # 1 F-Cl 0                                  1 F-Cl 0
        # sodium highlighted                        chloride and bond highlighted
        legendsMatrix = [["no highlighting", "bond highlighted"], 
        ["F highlighted", "", "Cl and bond highlighted"]]
        highlightAtomListsMatrix = [[[],[]], [[0], None, [1]]]
        highlightBondListsMatrix = [[[],[0]], [[], None, [0]]]

        dopts = rdMolDraw2D.MolDrawOptions()
        dopts.addAtomIndices = True

        img_binary = MolsMatrixToGridImage(molsMatrix=molsMatrix, subImgSize=(300, 400), 
        legendsMatrix=legendsMatrix, highlightAtomListsMatrix=highlightAtomListsMatrix, 
        highlightBondListsMatrix=highlightBondListsMatrix, useSVG=False, returnPNG=True, drawOptions=dopts)
        print(img_binary[:20])
        # Prints a binary string: b'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x03\x84'
  """
  mols, molsPerRow, legends, highlightAtomLists, highlightBondLists = _MolsNestedToLinear(
    molsMatrix, legendsMatrix, highlightAtomListsMatrix, highlightBondListsMatrix)

  return MolsToGridImage(mols, molsPerRow=molsPerRow, subImgSize=subImgSize, legends=legends,
                         highlightAtomLists=highlightAtomLists, useSVG=useSVG, returnPNG=returnPNG,
                         highlightBondLists=highlightBondLists, **kwargs)


def _MolsToGridSVG(mols, molsPerRow=3, subImgSize=(200, 200), legends=None, highlightAtomLists=None,
                   highlightBondLists=None, drawOptions=None, **kwargs):
  """ returns an SVG of the grid
  """
  if legends is None:
    legends = [''] * len(mols)

  nRows = len(mols) // molsPerRow
  if len(mols) % molsPerRow:
    nRows += 1

  blocks = [''] * (nRows * molsPerRow)

  fullSize = (molsPerRow * subImgSize[0], nRows * subImgSize[1])

  d2d = rdMolDraw2D.MolDraw2DSVG(fullSize[0], fullSize[1], subImgSize[0], subImgSize[1])
  if drawOptions is not None:
    d2d.SetDrawOptions(drawOptions)
  else:
    dops = d2d.drawOptions()
    for k, v in list(kwargs.items()):
      if hasattr(dops, k):
        setattr(dops, k, v)
        del kwargs[k]
  d2d.DrawMolecules(list(mols), legends=legends or None, highlightAtoms=highlightAtomLists or [],
                    highlightBonds=highlightBondLists or [], **kwargs)
  d2d.FinishDrawing()
  res = d2d.GetDrawingText()
  return res


def MolsToGridImage(mols, molsPerRow=3, subImgSize=(200, 200), legends=None,
                    highlightAtomLists=None, highlightBondLists=None, useSVG=False, returnPNG=False,
                    **kwargs):
  if legends and len(legends) > len(mols):
    legends = legends[:len(mols)]
  if highlightAtomLists and len(highlightAtomLists) > len(mols):
    highlightAtomLists = highlightAtomLists[:len(mols)]
  if highlightBondLists and len(highlightBondLists) > len(mols):
    highlightBondLists = highlightBondLists[:len(mols)]

  if useSVG:
    return _MolsToGridSVG(mols, molsPerRow=molsPerRow, subImgSize=subImgSize, legends=legends,
                          highlightAtomLists=highlightAtomLists,
                          highlightBondLists=highlightBondLists, **kwargs)
  else:
    return _MolsToGridImage(mols, molsPerRow=molsPerRow, subImgSize=subImgSize, legends=legends,
                            highlightAtomLists=highlightAtomLists,
                            highlightBondLists=highlightBondLists, returnPNG=returnPNG, **kwargs)


def ReactionToImage(rxn, subImgSize=(200, 200), useSVG=False, drawOptions=None, returnPNG=False,
                    **kwargs):
  if not useSVG and not hasattr(rdMolDraw2D, 'MolDraw2DCairo'):
    raise RuntimeError("PNG output requires that the RDKit be built with Cairo support")
  width = subImgSize[0] * (rxn.GetNumReactantTemplates() + rxn.GetNumProductTemplates() + 1)
  if useSVG:
    d = rdMolDraw2D.MolDraw2DSVG(width, subImgSize[1])
  else:
    d = rdMolDraw2D.MolDraw2DCairo(width, subImgSize[1])
  if drawOptions is not None:
    d.SetDrawOptions(drawOptions)
  d.DrawReaction(rxn, **kwargs)
  d.FinishDrawing()
  if useSVG or returnPNG:
    return d.GetDrawingText()
  else:
    return _drawerToImage(d)


def DebugDraw(mol, size=(350, 350), drawer=None, asSVG=True, useBW=True, includeHLabels=True,
              addAtomIndices=True, addBondIndices=False):
  if drawer is None:
    if asSVG:
      drawer = rdMolDraw2D.MolDraw2DSVG(size[0], size[1])
    else:
      drawer = rdMolDraw2D.MolDraw2DCairo(size[0], size[1])
  if useBW:
    drawer.drawOptions().useBWAtomPalette()

  drawer.drawOptions().addAtomIndices = addAtomIndices
  drawer.drawOptions().addBondIndices = addBondIndices

  drawer.drawOptions().annotationFontScale = 0.75

  if includeHLabels:
    for atom in mol.GetAtoms():
      if atom.GetTotalNumHs():
        atom.SetProp('atomNote', f'H{atom.GetTotalNumHs()}')

  aromAtoms = [x.GetIdx() for x in mol.GetAtoms() if x.GetIsAromatic()]
  clrs = {x: (.9, .9, .2) for x in aromAtoms}
  aromBonds = [x.GetIdx() for x in mol.GetBonds() if x.GetIsAromatic()]

  rdMolDraw2D.PrepareMolForDrawing(mol, kekulize=False, addChiralHs=False)
  drawer.drawOptions().prepareMolsBeforeDrawing = False

  drawer.DrawMolecule(mol, highlightAtoms=aromAtoms, highlightAtomColors=clrs,
                      highlightBonds=aromBonds)

  drawer.FinishDrawing()
  return drawer.GetDrawingText()


def DrawMorganBit(mol, bitId, bitInfo, whichExample=0, **kwargs):
  atomId, radius = bitInfo[bitId][whichExample]
  return DrawMorganEnv(mol, atomId, radius, **kwargs)


def DrawMorganBits(tpls, **kwargs):
  envs = []
  for tpl in tpls:
    if len(tpl) == 4:
      mol, bitId, bitInfo, whichExample = tpl
    else:
      mol, bitId, bitInfo = tpl
      whichExample = 0

    atomId, radius = bitInfo[bitId][whichExample]
    envs.append((mol, atomId, radius))
  return DrawMorganEnvs(envs, **kwargs)


# adapted from the function drawFPBits._drawFPBit() from the CheTo package
# original author Nadine Schneider
FingerprintEnv = namedtuple(
  'FingerprintEnv',
  ('submol', 'highlightAtoms', 'atomColors', 'highlightBonds', 'bondColors', 'highlightRadii'))


def _getMorganEnv(mol, atomId, radius, baseRad, aromaticColor, ringColor, centerColor, extraColor,
                  **kwargs):
  if not mol.GetNumConformers():
    rdDepictor.Compute2DCoords(mol)
  bitPath = Chem.FindAtomEnvironmentOfRadiusN(mol, radius, atomId)

  # get the atoms for highlighting
  atomsToUse = set((atomId, ))
  for b in bitPath:
    atomsToUse.add(mol.GetBondWithIdx(b).GetBeginAtomIdx())
    atomsToUse.add(mol.GetBondWithIdx(b).GetEndAtomIdx())

  #  enlarge the environment by one further bond
  enlargedEnv = set()
  for atom in atomsToUse:
    a = mol.GetAtomWithIdx(atom)
    for b in a.GetBonds():
      bidx = b.GetIdx()
      if bidx not in bitPath:
        enlargedEnv.add(bidx)
  enlargedEnv = list(enlargedEnv)
  enlargedEnv += bitPath

  # set the coordinates of the submol based on the coordinates of the original molecule
  amap = {}
  if enlargedEnv:
    submol = Chem.PathToSubmol(mol, enlargedEnv, atomMap=amap)
  else:
    # generate submol from fragments with no bonds
    submol = Chem.MolFromSmiles(Chem.MolFragmentToSmiles(mol, atomsToUse=atomsToUse))
  Chem.FastFindRings(submol)
  conf = Chem.Conformer(submol.GetNumAtoms())
  confOri = mol.GetConformer(0)
  for i1, i2 in amap.items():
    conf.SetAtomPosition(i2, confOri.GetAtomPosition(i1))
  submol.AddConformer(conf)
  envSubmol = []
  for i1, i2 in amap.items():
    for b in bitPath:
      beginAtom = amap[mol.GetBondWithIdx(b).GetBeginAtomIdx()]
      endAtom = amap[mol.GetBondWithIdx(b).GetEndAtomIdx()]
      envSubmol.append(submol.GetBondBetweenAtoms(beginAtom, endAtom).GetIdx())

  # color all atoms of the submol in gray which are not part of the bit
  # highlight atoms which are in rings
  atomcolors, bondcolors = {}, {}
  highlightAtoms, highlightBonds = [], []
  highlightRadii = {}
  for aidx in amap.keys():
    if aidx in atomsToUse:
      color = None
      if centerColor and aidx == atomId:
        color = centerColor
      elif aromaticColor and mol.GetAtomWithIdx(aidx).GetIsAromatic():
        color = aromaticColor
      elif ringColor and mol.GetAtomWithIdx(aidx).IsInRing():
        color = ringColor
      if color is not None:
        atomcolors[amap[aidx]] = color
        highlightAtoms.append(amap[aidx])
        highlightRadii[amap[aidx]] = baseRad
    else:
      #drawopt.atomLabels[amap[aidx]] = '*'
      submol.GetAtomWithIdx(amap[aidx]).SetAtomicNum(0)
      submol.GetAtomWithIdx(amap[aidx]).UpdatePropertyCache()
  color = extraColor
  for bid in submol.GetBonds():
    bidx = bid.GetIdx()
    if bidx not in envSubmol:
      bondcolors[bidx] = color
      highlightBonds.append(bidx)
  return FingerprintEnv(submol, highlightAtoms, atomcolors, highlightBonds, bondcolors,
                        highlightRadii)


def DrawMorganEnvs(envs, molsPerRow=3, subImgSize=(150, 150), baseRad=0.3, useSVG=True,
                   aromaticColor=(0.9, 0.9, 0.2), ringColor=(0.8, 0.8, 0.8),
                   centerColor=(0.6, 0.6, 0.9), extraColor=(0.9, 0.9, 0.9), legends=None,
                   drawOptions=None, **kwargs):
  submols = []
  highlightAtoms = []
  atomColors = []
  highlightBonds = []
  bondColors = []
  highlightRadii = []
  for mol, atomId, radius in envs:
    menv = _getMorganEnv(mol, atomId, radius, baseRad, aromaticColor, ringColor, centerColor,
                         extraColor, **kwargs)
    submols.append(menv.submol)
    highlightAtoms.append(menv.highlightAtoms)
    atomColors.append(menv.atomColors)
    highlightBonds.append(menv.highlightBonds)
    bondColors.append(menv.bondColors)
    highlightRadii.append(menv.highlightRadii)

  if legends is None:
    legends = [''] * len(envs)

  nRows = len(envs) // molsPerRow
  if len(envs) % molsPerRow:
    nRows += 1

  fullSize = (molsPerRow * subImgSize[0], nRows * subImgSize[1])
  # Drawing
  if useSVG:
    drawer = rdMolDraw2D.MolDraw2DSVG(fullSize[0], fullSize[1], subImgSize[0], subImgSize[1])
  else:
    drawer = rdMolDraw2D.MolDraw2DCairo(fullSize[0], fullSize[1], subImgSize[0], subImgSize[1])

  if drawOptions is None:
    drawOptions = drawer.drawOptions()
  drawOptions.continuousHighlight = False
  drawOptions.includeMetadata = False
  drawer.SetDrawOptions(drawOptions)
  drawer.DrawMolecules(submols, legends=legends, highlightAtoms=highlightAtoms,
                       highlightAtomColors=atomColors, highlightBonds=highlightBonds,
                       highlightBondColors=bondColors, highlightAtomRadii=highlightRadii, **kwargs)
  drawer.FinishDrawing()
  return drawer.GetDrawingText()


def DrawMorganEnv(mol, atomId, radius, molSize=(150, 150), baseRad=0.3, useSVG=True,
                  aromaticColor=(0.9, 0.9, 0.2), ringColor=(0.8, 0.8, 0.8),
                  centerColor=(0.6, 0.6, 0.9), extraColor=(0.9, 0.9, 0.9), drawOptions=None,
                  **kwargs):
  menv = _getMorganEnv(mol, atomId, radius, baseRad, aromaticColor, ringColor, centerColor,
                       extraColor, **kwargs)

  # Drawing
  if useSVG:
    drawer = rdMolDraw2D.MolDraw2DSVG(molSize[0], molSize[1])
  else:
    drawer = rdMolDraw2D.MolDraw2DCairo(molSize[0], molSize[1])

  if drawOptions is None:
    drawOptions = drawer.drawOptions()
  drawOptions.continuousHighlight = False
  drawOptions.includeMetadata = False
  drawer.SetDrawOptions(drawOptions)
  drawer.DrawMolecule(menv.submol, highlightAtoms=menv.highlightAtoms,
                      highlightAtomColors=menv.atomColors, highlightBonds=menv.highlightBonds,
                      highlightBondColors=menv.bondColors, highlightAtomRadii=menv.highlightRadii,
                      **kwargs)
  drawer.FinishDrawing()
  return drawer.GetDrawingText()


def DrawRDKitBits(tpls, **kwargs):
  envs = []
  for tpl in tpls:
    if len(tpl) == 4:
      mol, bitId, bitInfo, whichExample = tpl
    else:
      mol, bitId, bitInfo = tpl
      whichExample = 0

    bondpath = bitInfo[bitId][whichExample]
    envs.append((mol, bondpath))
  return DrawRDKitEnvs(envs, **kwargs)


def DrawRDKitBit(mol, bitId, bitInfo, whichExample=0, **kwargs):
  bondPath = bitInfo[bitId][whichExample]
  return DrawRDKitEnv(mol, bondPath, **kwargs)


def _getRDKitEnv(mol, bondPath, baseRad, aromaticColor, extraColor, nonAromaticColor, **kwargs):
  if not mol.GetNumConformers():
    rdDepictor.Compute2DCoords(mol)

  # get the atoms for highlighting
  atomsToUse = set()
  for b in bondPath:
    atomsToUse.add(mol.GetBondWithIdx(b).GetBeginAtomIdx())
    atomsToUse.add(mol.GetBondWithIdx(b).GetEndAtomIdx())

  # set the coordinates of the submol based on the coordinates of the original molecule
  amap = {}
  submol = Chem.PathToSubmol(mol, bondPath, atomMap=amap)
  Chem.FastFindRings(submol)
  conf = Chem.Conformer(submol.GetNumAtoms())
  confOri = mol.GetConformer(0)
  for i1, i2 in amap.items():
    conf.SetAtomPosition(i2, confOri.GetAtomPosition(i1))
  submol.AddConformer(conf)
  envSubmol = []
  for i1, i2 in amap.items():
    for b in bondPath:
      beginAtom = amap[mol.GetBondWithIdx(b).GetBeginAtomIdx()]
      endAtom = amap[mol.GetBondWithIdx(b).GetEndAtomIdx()]
      envSubmol.append(submol.GetBondBetweenAtoms(beginAtom, endAtom).GetIdx())

  # color all atoms of the submol in gray which are not part of the bit
  # highlight atoms which are in rings
  atomcolors, bondcolors = {}, {}
  highlightAtoms, highlightBonds = [], []
  highlightRadii = {}
  for aidx in amap.keys():
    if aidx in atomsToUse:
      color = None
      if aromaticColor and mol.GetAtomWithIdx(aidx).GetIsAromatic():
        color = aromaticColor
      elif nonAromaticColor and not mol.GetAtomWithIdx(aidx).GetIsAromatic():
        color = nonAromaticColor
      if color is not None:
        atomcolors[amap[aidx]] = color
        highlightAtoms.append(amap[aidx])
        highlightRadii[amap[aidx]] = baseRad
  color = extraColor
  for bid in submol.GetBonds():
    bidx = bid.GetIdx()
    if bidx not in envSubmol:
      bondcolors[bidx] = color
      highlightBonds.append(bidx)
  return FingerprintEnv(submol, highlightAtoms, atomcolors, highlightBonds, bondcolors,
                        highlightRadii)


def DrawRDKitEnvs(envs, molsPerRow=3, subImgSize=(150, 150), baseRad=0.3, useSVG=True,
                  aromaticColor=(0.9, 0.9, 0.2), extraColor=(0.9, 0.9, 0.9), nonAromaticColor=None,
                  legends=None, drawOptions=None, **kwargs):
  submols = []
  highlightAtoms = []
  atomColors = []
  highlightBonds = []
  bondColors = []
  highlightRadii = []
  for mol, bondpath in envs:
    menv = _getRDKitEnv(mol, bondpath, baseRad, aromaticColor, extraColor, nonAromaticColor,
                        **kwargs)
    submols.append(menv.submol)
    highlightAtoms.append(menv.highlightAtoms)
    atomColors.append(menv.atomColors)
    highlightBonds.append(menv.highlightBonds)
    bondColors.append(menv.bondColors)
    highlightRadii.append(menv.highlightRadii)

  if legends is None:
    legends = [''] * len(envs)

  nRows = len(envs) // molsPerRow
  if len(envs) % molsPerRow:
    nRows += 1

  fullSize = (molsPerRow * subImgSize[0], nRows * subImgSize[1])
  # Drawing
  if useSVG:
    drawer = rdMolDraw2D.MolDraw2DSVG(fullSize[0], fullSize[1], subImgSize[0], subImgSize[1])
  else:
    drawer = rdMolDraw2D.MolDraw2DCairo(fullSize[0], fullSize[1], subImgSize[0], subImgSize[1])

  if drawOptions is None:
    drawOptions = drawer.drawOptions()
  drawOptions.continuousHighlight = False
  drawOptions.includeMetadata = False
  drawer.SetDrawOptions(drawOptions)
  drawer.DrawMolecules(submols, legends=legends, highlightAtoms=highlightAtoms,
                       highlightAtomColors=atomColors, highlightBonds=highlightBonds,
                       highlightBondColors=bondColors, highlightAtomRadii=highlightRadii, **kwargs)
  drawer.FinishDrawing()
  return drawer.GetDrawingText()


def DrawRDKitEnv(mol, bondPath, molSize=(150, 150), baseRad=0.3, useSVG=True,
                 aromaticColor=(0.9, 0.9, 0.2), extraColor=(0.9, 0.9, 0.9), nonAromaticColor=None,
                 drawOptions=None, **kwargs):
  menv = _getRDKitEnv(mol, bondPath, baseRad, aromaticColor, extraColor, nonAromaticColor, **kwargs)

  # Drawing
  if useSVG:
    drawer = rdMolDraw2D.MolDraw2DSVG(molSize[0], molSize[1])
  else:
    drawer = rdMolDraw2D.MolDraw2DCairo(molSize[0], molSize[1])

  if drawOptions is None:
    drawOptions = drawer.drawOptions()
  drawOptions.continuousHighlight = False
  drawOptions.includeMetadata = False
  drawer.SetDrawOptions(drawOptions)
  drawer.DrawMolecule(menv.submol, highlightAtoms=menv.highlightAtoms,
                      highlightAtomColors=menv.atomColors, highlightBonds=menv.highlightBonds,
                      highlightBondColors=menv.bondColors, highlightAtomRadii=menv.highlightRadii,
                      **kwargs)
  drawer.FinishDrawing()
  return drawer.GetDrawingText()


def SetComicMode(opts):
  opts.fontFile = os.path.join(RDConfig.RDDataDir, "Fonts", "ComicNeue-Regular.ttf")
  opts.comicMode = True


def _getMatchAtomsAndBonds(mol, qry, match):
  alist = match
  blist = []
  if qry is not None:
    for bnd in qry.GetBonds():
      b = mol.GetBondBetweenAtoms(alist[bnd.GetBeginAtomIdx()], alist[bnd.GetEndAtomIdx()])
      blist.append(b.GetIdx())
  return alist, blist


def _addColorsToMap(els, cols, color):
  for el in els:
    if el not in cols:
      cols[el] = []
    if color not in cols[el]:
      cols[el].append(color)


def DrawMolWithMatches(mol, matches, molSize=(350, 300), qry=None, label='',
                       options=MolDrawOptions(), colors=None, doPNG=True, kekulize=True, confId=-1):
  ''' Draws a molecule with a set of substructure matches highlighted

      ARGUMENTS:

        - mol: The molecule to draw

        - matches: A list/tuple of lists/tuples of atom indices, each inner tuple representing a substructure match

        - molSize: The size of the image to produce (default (350, 300))

        - qry: The query molecule used to generate the matches (default None)
                If supplied, bonds between matched atoms will also be highlighted

        - label: A string to place under the molecule (default '')

        - options: A MolDrawOptions object to control the drawing style (default MolDrawOptions())

        - colors: A list/tuple of colors to use for highlighting the different matches
                  (default None, in which case the default highlight color is used for all matches)

        - doPNG: Whether to return PNG (if true) or SVG (if false) (default true)

        - kekulize: Whether to kekulize the molecule before drawing (default true)

        - confId: The conformer ID to use when drawing (default -1, the default conformer)

      RETURNS:
        A string containing the image in the requested format (PNG or SVG)
'''
  mol = _prepareMol(mol, options, kekulize=kekulize)

  acols = {}
  bcols = {}
  h_rads = {}
  h_lw_mult = {}
  for i, match in enumerate(matches):
    alist, blist = _getMatchAtomsAndBonds(mol, qry, match)
    for ai in alist:
      h_rads[ai] = options.highlightRadius
    for bi in blist:
      h_lw_mult[bi] = options.highlightBondWidthMultiplier
    _addColorsToMap(alist, acols,
                    colors[i % len(colors)] if colors is not None else options.getHighlightColour())
    _addColorsToMap(blist, bcols,
                    colors[i % len(colors)] if colors is not None else options.getHighlightColour())
  d = MolDraw2DCairo(molSize[0], molSize[1]) if doPNG else MolDraw2DSVG(molSize[0], molSize[1])
  d.SetDrawOptions(options)
  d.drawOptions().prepareMolsBeforeDrawing = False
  d.DrawMoleculeWithHighlights(mol, label, acols, bcols, h_rads, h_lw_mult, confId=confId)
  d.FinishDrawing()
  return d.GetDrawingText()
